from __future__ import annotations

import json
import os
from pathlib import Path
from shutil import copytree, ignore_patterns
from tempfile import gettempdir
from typing import Any, Dict, List, Union

import click
from loguru import logger
from pydantic import BaseSettings, Extra, Field, PydanticValueError, validator

from .utils import Literal, trash

__novalidate__ = '__novalidate__'


class _PathNotWritableDirectoryError(PydanticValueError):
    code = 'path.not_writable_directory'
    msg_template = '"{path}" is not a writable directory'


def _get_default_config_dir() -> Path:
    return Path(click.get_app_dir('instawow'))


def _validate_expand_path(value: Path) -> Path:
    try:
        return value.expanduser().resolve()
    except RuntimeError as error:
        # pathlib will raise a ``RuntimeError`` for non-existent ~users
        raise ValueError(str(error)) from error


def _validate_path_is_writable_dir(value: Path) -> Path:
    if not (value.is_dir() and os.access(value, os.W_OK)):
        raise _PathNotWritableDirectoryError(path=value)

    return value


class BaseConfig(BaseSettings):
    def _build_values(
        self,
        init_kwargs: Dict[str, Any],
        *args: Any,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        # Prioritise env vars
        return {**init_kwargs, **self._build_environ()}


class _GlobalConfig(BaseConfig):
    config_dir: Path = Field(default_factory=_get_default_config_dir)
    profile: str = Field('__default__', min_length=1, strip_whitespace=True)
    addon_dir: Path
    game_flavour: Literal['retail', 'classic']
    auto_update_check: bool = True
    temp_dir: Path = Path(gettempdir()) / 'instawow'

    class Config:  # type: ignore
        env_prefix = 'INSTAWOW_'
        extra = Extra.allow
        validate_assignment = True

    @validator('config_dir', 'addon_dir', 'temp_dir')
    def _validate_expand_path(cls, value: Path) -> Path:
        return _validate_expand_path(value)

    @validator('addon_dir')
    def _validate_path_is_writable_dir(cls, value: Path) -> Path:
        if value.name == __novalidate__:
            return value
        return _validate_path_is_writable_dir(value)

    @classmethod
    def get_dummy_config(cls, **kwargs: Any) -> _GlobalConfig:
        "Create a dummy configuration with default values."
        template = {'game_flavour': 'retail', 'addon_dir': __novalidate__}
        dummy_config = cls.parse_obj({**template, **kwargs})
        return dummy_config

    @classmethod
    def list_profiles(cls) -> List[str]:
        "List the profiles contained in the default ``config_dir``."
        dummy_config = cls.get_dummy_config()
        profiles = [f.name for f in (dummy_config.config_dir / 'profiles').iterdir() if f.is_dir()]
        return profiles

    @classmethod
    def read(cls, profile: str) -> _GlobalConfig:
        "Read the configuration from disk."
        dummy_config = cls.get_dummy_config(profile=profile)
        dummy_config.migrate_legacy_dirs()
        config = cls.parse_raw(dummy_config.config_file.read_text(encoding='utf-8'))
        if dummy_config.profile != config.profile:
            raise ValueError(
                'profile location does not match profile value of '
                f'"{config.profile}" in {dummy_config.config_file}'
            )
        return config

    def ensure_dirs(self) -> _GlobalConfig:
        "Create the various folders used by instawow."
        for dir_ in (
            self.config_dir,
            self.profile_dir,
            self.logging_dir,
            self.plugin_dir,
            self.temp_dir,
            self.cache_dir,
        ):
            dir_.mkdir(exist_ok=True, parents=True)
        return self

    def write(self) -> _GlobalConfig:
        """Write the configuration on disk.

        ``write``, unlike ``ensure_dirs``, should only be called when configuring
        instawow.  This means that environment overrides should only be persisted
        if made during configuration.
        """
        self.ensure_dirs()
        includes = {'addon_dir', 'game_flavour', 'profile'}
        output = self.json(include=includes, indent=2)
        self.config_file.write_text(output, encoding='utf-8')
        return self

    def migrate_legacy_dirs(self) -> _GlobalConfig:
        "Migrate a profile-less configuration to the new format."
        legacy_config_file = self.config_dir / 'config.json'
        if (
            self.profile == self.__field_defaults__['profile']
            and not self.profile_dir.exists()
            and legacy_config_file.exists()
        ):
            legacy_json = json.loads(legacy_config_file.read_text(encoding='utf-8'))
            legacy_json.pop('profile', None)
            legacy_config = self.parse_obj(legacy_json)
            ignores = ignore_patterns('profiles')

            logger.info('migrating legacy configuration')
            copytree(self.config_dir, self.profile_dir, ignore=ignores)
            legacy_config.write()
            trash(
                [i for i in self.config_dir.iterdir() if i.name != 'profiles'], dst=self.temp_dir
            )

        return self

    def delete(self) -> None:
        "Delete the configuration files associated with this profile."
        trash((self.ensure_dirs().profile_dir,), dst=self.temp_dir, missing_ok=True)

    @property
    def is_classic(self) -> bool:
        return self.game_flavour == 'classic'

    @property
    def is_retail(self) -> bool:
        return self.game_flavour == 'retail'

    @property
    def profile_dir(self) -> Path:
        return self.config_dir / 'profiles' / self.profile

    @property
    def logging_dir(self) -> Path:
        return self.profile_dir / 'logs'

    @property
    def plugin_dir(self) -> Path:
        return self.profile_dir / 'plugins'

    @property
    def config_file(self) -> Path:
        return self.profile_dir / 'config.json'

    @property
    def db_file(self) -> Path:
        return self.profile_dir / 'db.sqlite'

    @property
    def cache_dir(self) -> Path:
        return self.temp_dir / 'cache'


Config = _GlobalConfig


def setup_logging(config: _GlobalConfig, log_level: Union[int, str] = 'INFO') -> int:
    import logging

    class InterceptHandler(logging.Handler):
        def emit(self, record: logging.LogRecord):
            # Get the corresponding Loguru level if it exists
            try:
                level = logger.level(record.levelname).name
            except ValueError:
                level = record.levelno

            # Find caller from where the logged message originated
            frame = logging.currentframe()
            depth = 2
            while frame and frame.f_code.co_filename == getattr(logging, '__file__', None):
                frame = frame.f_back
                depth += 1

            logger.opt(
                depth=depth,
                exception=record.exc_info,  # type: ignore
            ).log(level, record.getMessage())

    logging.basicConfig(handlers=[InterceptHandler()], level=log_level)
    handler = {
        'sink': config.logging_dir / 'error.log',
        'level': log_level,
        'rotation': '1 MB',
        'enqueue': True,
    }
    (handler_id,) = logger.configure(handlers=(handler,))
    return handler_id



from __future__ import annotations

from datetime import datetime
import enum
from itertools import chain, count, takewhile
import re
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    ClassVar,
    Dict,
    List,
    Optional as O,
    Sequence,
    Set,
    Tuple,
    cast,
)

from pydantic import BaseModel
from yarl import URL

from . import exceptions as E, models as m
from .utils import Literal, cached_property, gather, run_in_thread as t, slugify, uniq

if TYPE_CHECKING:
    from .manager import Manager

    JsonDict = Dict[str, Any]


class Strategies(enum.Enum):
    default = 'default'
    latest = 'latest'
    curse_latest_beta = 'curse_latest_beta'
    curse_latest_alpha = 'curse_latest_alpha'
    any_flavour = 'any_flavour'
    version = 'version'


class Defn(BaseModel):
    source: str
    source_id: O[str] = None
    name: str
    strategy: Strategies = Strategies.default
    strategy_vals: Tuple[str, ...] = ()

    @classmethod
    def from_pkg(cls, pkg: m.Pkg) -> Defn:
        defn = cls(
            source=pkg.source, source_id=pkg.id, name=pkg.slug, strategy=pkg.options.strategy
        )
        if defn.strategy is Strategies.version:
            defn.strategy_vals = (pkg.version,)
        return defn

    @classmethod
    def get(cls, source: str, name: str) -> Defn:
        return cls(source=source, name=name)

    def with_(self, **kwargs: Any) -> Defn:
        return self.__class__(**{**self.__dict__, **kwargs})

    def with_version(self, version: str) -> Defn:
        return self.with_(strategy='version', strategy_vals=(version,))

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            other = tuple(other.__dict__.values())
        return tuple(self.__dict__.values()) == other

    def __hash__(self) -> int:
        return hash(tuple(self.__dict__.values()))

    def __str__(self) -> str:
        return f'{self.source}:{self.name}'


class _CatalogueEntry(BaseModel):
    source: str
    id: str
    slug: str = ''
    name: str
    compatibility: Set[Literal['retail', 'classic']] = {'retail', 'classic'}
    folders: List[List[str]] = []


class MasterCatalogue(BaseModel):
    __root__: List[_CatalogueEntry]

    class Config:
        keep_untouched = (cast(Any, cached_property),)

    @classmethod
    async def collate(cls) -> MasterCatalogue:
        from types import SimpleNamespace

        from .manager import init_web_client

        resolvers = (CurseResolver, WowiResolver, TukuiResolver)

        async with init_web_client() as web_client:
            faux_manager = cast('Manager', SimpleNamespace(web_client=web_client))
            items = [a for r in resolvers async for a in r(faux_manager).collect_items()]
        catalogue = cls(__root__=items)
        return catalogue

    @cached_property
    def curse_slugs(self) -> Dict[str, str]:
        return {a.slug: a.id for a in self.__root__ if a.source == 'curse'}


class Resolver:
    source: ClassVar[str]
    name: ClassVar[str]
    strategies: ClassVar[Set[Strategies]]

    def __init__(self, manager: Manager) -> None:
        self.manager = manager

    def __init_subclass__(cls) -> None:
        async def resolve_wrapper(self: Resolver, defn: Defn, metadata: O[JsonDict]) -> m.Pkg:
            if defn.strategy in self.strategies:
                return await resolve_one(self, defn, metadata)  # type: ignore
            raise E.PkgStrategyUnsupported(defn.strategy)

        resolve_one = cls.resolve_one
        cls.resolve_one = resolve_wrapper  # type: ignore

    @property
    def supports_rollback(self) -> bool:
        "Whether the resolver supports rollback operations."
        return Strategies.version in self.strategies

    @staticmethod
    def get_name_from_url(value: str) -> O[str]:
        "Attempt to extract a definition name from a given URL."

    async def resolve(self, defns: Sequence[Defn]) -> Dict[Defn, Any]:
        "Resolve add-on definitions into packages."
        results = await gather(self.resolve_one(d, None) for d in defns)
        return dict(zip(defns, results))

    async def resolve_one(self, defn: Defn, metadata: O[JsonDict]) -> m.Pkg:
        "Resolve an individual definition into a package."
        raise NotImplementedError

    async def collect_items(self) -> AsyncIterable[_CatalogueEntry]:
        "Yield add-ons from source for cataloguing."
        return
        yield


class CurseResolver(Resolver):
    source = 'curse'
    name = 'CurseForge'
    strategies = {
        Strategies.default,
        Strategies.latest,
        Strategies.curse_latest_beta,
        Strategies.curse_latest_alpha,
        Strategies.any_flavour,
        Strategies.version,
    }

    # Reference: https://twitchappapi.docs.apiary.io/
    addon_api_url = URL('https://addons-ecs.forgesvc.net/api/v2/addon')

    @staticmethod
    def get_name_from_url(value: str) -> O[str]:
        url = URL(value)
        if url.host == 'www.wowace.com' and len(url.parts) > 2 and url.parts[1] == 'projects':
            return url.parts[2].lower()
        elif (
            url.host == 'www.curseforge.com'
            and len(url.parts) > 3
            and url.parts[1:3] == ('wow', 'addons')
        ):
            return url.parts[3].lower()

    async def resolve(self, defns: Sequence[Defn]) -> Dict[Defn, Any]:
        from aiohttp import ClientResponseError

        from .manager import cache_json_response

        ids_for_defns = {
            d: (d.source_id or self.manager.catalogue.curse_slugs.get(d.name) or d.name)
            for d in defns
        }
        numeric_ids = {i for i in ids_for_defns.values() if i.isdigit()}
        try:
            json_response = await cache_json_response(
                self.manager,
                self.addon_api_url,
                300,
                request_kwargs={'method': 'POST', 'json': list(numeric_ids)},
            )
        except ClientResponseError as error:
            if error.status != 404:
                raise
            json_response = []

        api_results = {str(r['id']): r for r in json_response}
        results = await gather(
            self.resolve_one(d, api_results.get(i)) for d, i in ids_for_defns.items()
        )
        return dict(zip(defns, results))

    async def resolve_one(self, defn: Defn, metadata: O[JsonDict]) -> m.Pkg:
        if not metadata:
            raise E.PkgNonexistent

        if defn.strategy is Strategies.version:

            from .manager import cache_json_response

            files = await cache_json_response(
                self.manager,
                self.addon_api_url / str(metadata['id']) / 'files',
                3600,
                label=f'Fetching metadata from {self.name}',
            )

            def is_compatible(f: JsonDict):
                return defn.strategy_vals[0] == f['displayName']

        else:

            files = metadata['latestFiles']

            def is_not_libless(f: JsonDict):
                # There's also an 'isAlternate' field that's missing from some
                # 50 lib-less files from c. 2008.  'exposeAsAlternative' is
                # absent from the /file endpoint
                return not f['exposeAsAlternative']

            if defn.strategy is Strategies.any_flavour:

                def supports_game_version(f: JsonDict):
                    return True

            else:
                classic_version_prefix = '1.13'
                flavour = 'wow_classic' if self.manager.config.is_classic else 'wow_retail'

                def supports_game_version(f: JsonDict):
                    # Files can belong both to retail and classic
                    # but ``gameVersionFlavor`` can only be one of
                    # 'wow_retail' or 'wow_classic'.  To spice things up,
                    # ``gameVersion`` might not be populated so we still have to
                    # check the value of ``gameVersionFlavor``
                    return f['gameVersionFlavor'] == flavour or any(
                        v.startswith(classic_version_prefix) is self.manager.config.is_classic
                        for v in f['gameVersion']
                    )

            # 1 = stable; 2 = beta; 3 = alpha
            if defn.strategy is Strategies.latest:

                def has_release_type(f: JsonDict):
                    return True

            elif defn.strategy is Strategies.curse_latest_beta:

                def has_release_type(f: JsonDict):
                    return f['releaseType'] == 2

            elif defn.strategy is Strategies.curse_latest_alpha:

                def has_release_type(f: JsonDict):
                    return f['releaseType'] == 3

            else:

                def has_release_type(f: JsonDict):
                    return f['releaseType'] == 1

            def is_compatible(f: JsonDict):
                return is_not_libless(f) and supports_game_version(f) and has_release_type(f)

        if not files:
            raise E.PkgFileUnavailable('no files available for download')
        try:
            file = max(
                filter(is_compatible, files),
                # The ``id`` is just a counter so we don't have to go digging around dates
                key=lambda f: f['id'],
            )
        except ValueError:
            raise E.PkgFileUnavailable(
                f'no files compatible with {self.manager.config.game_flavour} '
                f'using {defn.strategy.name!r} strategy'
            )

        # 1 = embedded library
        # 2 = optional dependency
        # 3 = required dependency
        # 4 = tool
        # 5 = incompatible
        # 6 = include (wat)
        deps = [m.PkgDep(id=d['addonId']) for d in file['dependencies'] if d['type'] == 3]

        return m.Pkg(
            source=self.source,
            id=metadata['id'],
            slug=metadata['slug'],
            name=metadata['name'],
            description=metadata['summary'],
            url=metadata['websiteUrl'],
            download_url=file['downloadUrl'],
            date_published=file['fileDate'],
            version=file['displayName'],
            options=m.PkgOptions(strategy=defn.strategy.name),
            deps=deps,
        )

    async def collect_items(self) -> AsyncIterable[_CatalogueEntry]:
        classic_version_prefix = '1.13'
        flavours = ('retail', 'classic')

        def excise_compatibility(files: Any):
            for c in flavours:
                if any(f['gameVersionFlavor'] == f'wow_{c}' for f in files):
                    yield c
                elif any(
                    v.startswith(classic_version_prefix) is (c == 'classic')
                    for f in files
                    for v in f['gameVersion']
                ):
                    yield c

        step = 1000
        sort_order = '3'  # Alphabetical
        for index in count(0, step):
            url = (self.addon_api_url / 'search').with_query(
                gameId='1', sort=sort_order, pageSize=step, index=index
            )
            async with self.manager.web_client.get(url) as response:
                json_response = await response.json()

            if not json_response:
                break
            for item in json_response:
                folders = uniq(
                    tuple(m['foldername'] for m in f['modules'])
                    for f in item['latestFiles']
                    if not f['exposeAsAlternative']
                )
                yield _CatalogueEntry(
                    source=self.source,
                    id=item['id'],
                    slug=item['slug'],
                    name=item['name'],
                    compatibility=set(excise_compatibility(item['latestFiles'])),
                    folders=folders,
                )


class WowiResolver(Resolver):
    source = 'wowi'
    name = 'WoWInterface'
    strategies = {Strategies.default}

    # Reference: https://api.mmoui.com/v3/globalconfig.json
    # There's also a v4 API corresponding to the as yet unreleased Minion v4,
    # which is fair to assume is unstable.  They changed the naming scheme to
    # camelCase and some fields which were strings were converted to numbers.
    # Neither API provides access to classic files for multi-file add-ons and
    # 'UICompatibility' can't be relied on to enforce compatibility
    # in instawow.  The API appears to inherit the version of the _latest_
    # file to have been uploaded, which for multi-file add-ons can be the
    # classic version.  Hoooowever the download link always points to the
    # 'retail' version, which for single-file add-ons belonging to the
    # classic category would be an add-on for classic.
    list_api_url = 'https://api.mmoui.com/v3/game/WOW/filelist.json'
    details_api_url = URL('https://api.mmoui.com/v3/game/WOW/filedetails/')

    _files: O[JsonDict] = None

    @staticmethod
    def get_name_from_url(value: str) -> O[str]:
        url = URL(value)
        if (
            url.host in {'wowinterface.com', 'www.wowinterface.com'}
            and len(url.parts) == 3
            and url.parts[1] == 'downloads'
        ):
            if url.name == 'landing.php':
                source_id = url.query.get('fileid')
                if source_id:
                    return source_id
            elif url.name == 'fileinfo.php':
                source_id = url.query.get('id')
                if source_id:
                    return source_id
            else:
                match = re.match(r'^(?:download|info)(?P<id>\d+)', url.name)
                if match:
                    return match.group('id')

    async def resolve(self, defns: Sequence[Defn]) -> Dict[Defn, Any]:
        from aiohttp import ClientResponseError

        from .manager import cache_json_response

        async with self.manager.locks['load WoWI catalogue']:
            if self._files is None:
                files = await cache_json_response(
                    self.manager,
                    self.list_api_url,
                    3600,
                    label=f'Synchronising {self.name} catalogue',
                )
                self._files = {i['UID']: i for i in files}

        ids_for_defns = {d: ''.join(takewhile(str.isdigit, d.name)) for d in defns}
        numeric_ids = {i for i in ids_for_defns.values() if i.isdigit()}
        url = self.details_api_url / f'{",".join(numeric_ids)}.json'
        try:
            json_response = await cache_json_response(self.manager, url, 300)
        except ClientResponseError as error:
            if error.status != 404:
                raise
            json_response = []

        api_results = {r['UID']: {**self._files[r['UID']], **r} for r in json_response}
        results = await gather(
            self.resolve_one(d, api_results.get(i)) for d, i in ids_for_defns.items()
        )
        return dict(zip(defns, results))

    async def resolve_one(self, defn: Defn, metadata: O[JsonDict]) -> m.Pkg:
        if not metadata:
            raise E.PkgNonexistent

        # 'UIPending' is set to '1' for files awaiting approval during which
        # time 'UIMD5' is null - all other fields appear to be filled correctly
        return m.Pkg(
            source=self.source,
            id=metadata['UID'],
            slug=slugify(f'{metadata["UID"]} {metadata["UIName"]}'),
            name=metadata['UIName'],
            description=metadata['UIDescription'],
            url=metadata['UIFileInfoURL'],
            download_url=metadata['UIDownload'],
            date_published=metadata['UIDate'],
            version=metadata['UIVersion'],
            options=m.PkgOptions(strategy=defn.strategy.name),
        )

    async def collect_items(self) -> AsyncIterable[_CatalogueEntry]:
        async with self.manager.web_client.get(self.list_api_url) as response:
            json_response = await response.json()

        for item in json_response:
            yield _CatalogueEntry(
                source=self.source,
                id=item['UID'],
                name=item['UIName'],
                folders=[item['UIDir']],
            )


class TukuiResolver(Resolver):
    source = 'tukui'
    name = 'Tukui'
    strategies = {Strategies.default}

    api_url = URL('https://www.tukui.org/api.php')

    retail_uis = {'-1': 'tukui', '-2': 'elvui', 'tukui': 'tukui', 'elvui': 'elvui'}

    @staticmethod
    def get_name_from_url(value: str) -> O[str]:
        url = URL(value)
        if url.host == 'www.tukui.org' and url.path in {
            '/addons.php',
            '/classic-addons.php',
            '/download.php',
        }:
            name = url.query.get('id') or url.query.get('ui')
            if name:
                return name

    async def resolve_one(self, defn: Defn, metadata: Any) -> m.Pkg:
        name = ui_name = self.retail_uis.get(defn.name)
        if not name:
            name = ''.join(takewhile('-'.__ne__, defn.name))

        if self.manager.config.is_classic:
            query = 'classic-addon'
        elif ui_name:
            query = 'ui'
        else:
            query = 'addon'

        url = self.api_url.with_query({query: name})
        async with self.manager.web_client.get(url) as response:
            if not response.content_length:
                raise E.PkgNonexistent
            addon = await response.json(content_type=None)  # text/html

        return m.Pkg(
            source=self.source,
            id=addon['id'],
            slug=ui_name or slugify(f'{addon["id"]} {addon["name"]}'),
            name=addon['name'],
            description=addon['small_desc'],
            url=addon['web_url'],
            download_url=addon['url'],
            date_published=datetime.fromisoformat(addon['lastupdate']),
            version=addon['version'],
            options=m.PkgOptions(strategy=defn.strategy.name),
        )

    async def collect_items(self) -> AsyncIterable[_CatalogueEntry]:
        for query, param, compatibility in [
            ('ui', 'tukui', 'retail'),
            ('ui', 'elvui', 'retail'),
            ('addons', 'all', 'retail'),
            ('classic-addons', 'all', 'classic'),
        ]:
            url = self.api_url.with_query({query: param})
            async with self.manager.web_client.get(url) as response:
                metadata = await response.json(content_type=None)  # text/html

            if query == 'ui':
                yield _CatalogueEntry(
                    source=self.source,
                    id=metadata['id'],
                    name=metadata['name'],
                    compatibility={compatibility},
                )
            else:
                for item in metadata:
                    yield _CatalogueEntry(
                        source=self.source,
                        id=item['id'],
                        name=item['name'],
                        compatibility={compatibility},
                    )


class GithubResolver(Resolver):
    source = 'github'
    name = 'GitHub'
    strategies = {
        Strategies.default,
        Strategies.version,
    }

    repos_api_url = URL('https://api.github.com/repos')

    @staticmethod
    def get_name_from_url(value: str) -> O[str]:
        url = URL(value)
        if url.host == 'github.com' and len(url.parts) > 2:
            return '/'.join(url.parts[1:3])

    async def resolve_one(self, defn: Defn, metadata: Any) -> m.Pkg:
        """Resolve a hypothetical add-on hosted on GitHub.

        The GitHub resolver, ahem, 'builds' on the work done by Torkus
        (see https://github.com/ogri-la/strongbox/blob/develop/github-addons.md) -
        it purports to support the different kinds of add-ons supported by strongbox
        with the exception that instawow does not look for TOC files or validate
        the contents of the ZIP file.
        instawow will attempt to prioritise add-ons compatible with your selected
        game flavour.  It will otherwise install the first file it encounters.
        instawow will only install assets attached to releases.  It will not
        install add-ons from VCS tarballs or 'zipballs' (i.e. from source).
        """
        from aiohttp import ClientResponseError

        from .manager import cache_json_response

        repo_url = self.repos_api_url / defn.name
        try:
            project_metadata = await cache_json_response(self.manager, repo_url, 3600)
        except ClientResponseError as error:
            if error.status == 404:
                raise E.PkgNonexistent
            raise

        if defn.strategy is Strategies.version:
            release_url = repo_url / 'releases/tags' / defn.strategy_vals[0]
        else:
            release_url = repo_url / 'releases/latest'
        async with self.manager.web_client.get(release_url) as response:
            if response.status == 404:
                raise E.PkgFileUnavailable('release not found')
            release_metadata = await response.json()

        try:

            def is_valid_asset(asset: JsonDict):
                return (
                    # There is something of a convention that Classic archives
                    # end in '-classic' and lib-less archives end in '-nolib'.
                    # Archives produced by packager follow this convention
                    not asset['name'].endswith(('-nolib.zip', '-nolib-classic.zip'))
                    and asset['content_type']
                    in {'application/zip', 'application/x-zip-compressed'}
                    # A failed upload has a state value of 'starter'
                    and asset['state'] == 'uploaded'
                )

            assets = release_metadata.get('assets', [])
            matching_asset = next(
                chain(
                    (
                        a
                        for a in assets
                        if is_valid_asset(a)
                        and a['name'].endswith('-classic.zip') is self.manager.config.is_classic
                    ),
                    filter(is_valid_asset, assets),
                )
            )
        except StopIteration:
            raise E.PkgFileUnavailable

        return m.Pkg(
            source=self.source,
            id=project_metadata['full_name'],
            slug=project_metadata['full_name'].lower(),
            name=project_metadata['name'],
            description=project_metadata['description'],
            url=project_metadata['html_url'],
            download_url=matching_asset['browser_download_url'],
            date_published=release_metadata['published_at'],
            version=release_metadata['tag_name'],
            options=m.PkgOptions(strategy=defn.strategy.name),
        )


class InstawowResolver(Resolver):
    source = 'instawow'
    name = 'instawow'
    strategies = {Strategies.default}

    _addons = {
        ('0', 'weakauras-companion'),
        ('1', 'weakauras-companion-autoupdate'),
    }

    async def resolve_one(self, defn: Defn, metadata: Any) -> m.Pkg:
        try:
            source_id, slug = next(p for p in self._addons if defn.name in p)
        except StopIteration:
            raise E.PkgNonexistent

        from .wa_updater import BuilderConfig, WaCompanionBuilder

        sentinel = '__sentinel__'
        builder = WaCompanionBuilder(self.manager, BuilderConfig(account=sentinel))
        if source_id == '1':
            if builder.builder_config.account == sentinel:
                raise E.PkgFileUnavailable('account name not provided')
            await builder.build()

        checksum = await t(builder.checksum)()
        return m.Pkg(
            source=self.source,
            id=source_id,
            slug=slug,
            name='WeakAuras Companion',
            description='A WeakAuras Companion clone.',
            url='https://github.com/layday/instawow',
            download_url=builder.addon_file.as_uri(),
            date_published=datetime.now(),
            version=checksum[:7],
            options=m.PkgOptions(strategy=defn.strategy.name),
        )



from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional as O, Sequence, Tuple, Type

from loguru import logger
from pydantic import BaseModel, Field, validator
from yarl import URL

from .config import BaseConfig, Config as GlobalConfig
from .utils import bucketise, chain_dict, gather, iter_in_thread as iit, run_in_thread as t

if TYPE_CHECKING:
    from pathlib import Path

    from .manager import Manager

    ImportString = str
    RemoteAura = Tuple[List['WeakAura'], 'WagoResponse', ImportString]


import_api = URL('https://data.wago.io/api/raw/encoded')


class BuilderConfig(BaseConfig):
    account: str
    api_key: O[str]

    class Config:  # type: ignore
        env_prefix = 'WAC_'

    def get_saved_vars(self, global_config: GlobalConfig) -> Path:
        root_for_flavour = global_config.addon_dir.parents[1]
        return root_for_flavour / 'WTF' / 'Account' / self.account / 'SavedVariables'


class WeakAura(BaseModel):
    id: str
    uid: str
    parent: O[str]
    url: URL
    version: int

    class Config:
        arbitrary_types_allowed = True

    @validator('url', pre=True)
    def _convert_url(cls, value: str) -> URL:
        return URL(value)


class WeakAuras(BaseModel):
    entries: Dict[str, List[WeakAura]]

    class Config:
        arbitrary_types_allowed = True

    class Meta:
        model = WeakAura
        filename = 'WeakAuras.lua'
        table_prefix = 'WeakAurasSaved'
        api = URL('https://data.wago.io/api/check/weakauras')

    @classmethod
    def from_lua_table(cls, lua_table: Dict[Any, Any]) -> WeakAuras:
        auras = (
            cls.Meta.model.parse_obj(a) for a in lua_table['displays'].values() if a.get('url')
        )
        return cls(entries=bucketise(auras, key=lambda a: a.url.parts[1]))


class Plateroo(WeakAura):
    id: str = Field(alias='Name')
    uid = ''


class Plateroos(WeakAuras):
    class Meta:  # type: ignore
        model = Plateroo
        filename = 'Plater.lua'
        table_prefix = 'PlaterDB'
        api = URL('https://data.wago.io/api/check/plater')

    @classmethod
    def from_lua_table(cls, lua_table: Dict[Any, Any]) -> Plateroos:
        auras = (
            cls.Meta.model.parse_obj(a)
            for n, p in lua_table['profiles'].items()
            for a in chain(
                ({**p, 'Name': f'__profile_{n}__'},),
                (i for n, v in p.items() if n in {'script_data', 'hook_data'} for i in v),
            )
            if a.get('url')
        )
        return cls(entries={a.url.parts[1]: [a] for a in auras})


class _WagoChangelog(BaseModel):
    format: O[str]
    text: str = ''


class WagoResponse(BaseModel):
    id: str = Field(alias='_id')
    name: str  # +
    slug: str  # +
    url: str
    created: str
    modified: str
    game: str
    fork_of: O[str] = Field(alias='forkOf')
    username: str  # +
    version: int  # +
    version_string: str = Field(alias='versionString')  # +
    changelog: _WagoChangelog  # +
    region_type: O[str] = Field(alias='regionType')


class WaCompanionBuilder:
    """A WeakAuras Companion port for shellfolk."""

    def __init__(self, manager: Manager, builder_config: BuilderConfig) -> None:
        self.manager = manager
        self.addon_file = self.manager.config.plugin_dir / __name__ / 'WeakAurasCompanion.zip'
        self.builder_config = builder_config

    @staticmethod
    def extract_auras(model: Type[WeakAuras], source: str) -> WeakAuras:
        import re

        from slpp import SLPP

        class WaParser(SLPP):
            def decode(self, text: str) -> Any:
                text = re.sub(rf'^\s*{model.Meta.table_prefix} = ', '', text)
                text = re.sub(r' -- \[\d+\]$', '', text, flags=re.M)
                self.text = text
                self.at, self.ch, self.depth = 0, '', 0
                self.len = len(text)
                self.next_chr()
                return self.value()  # type: ignore

        table = WaParser().decode(source)
        return model.from_lua_table(table)

    def extract_installed_auras(self) -> Iterator[WeakAuras]:
        import time

        saved_vars = self.builder_config.get_saved_vars(self.manager.config)
        for model in WeakAuras, Plateroos:
            file = saved_vars / model.Meta.filename
            if not file.exists():
                logger.info(f'{file} not found')
            else:
                start = time.perf_counter()
                aura_groups = self.extract_auras(
                    model, file.read_text(encoding='utf-8-sig', errors='replace')
                )
                logger.debug(f'{model.__name__} extracted in {time.perf_counter() - start}s')
                yield aura_groups

    async def get_wago_metadata(self, aura_groups: WeakAuras) -> List[WagoResponse]:
        aura_ids = list(aura_groups.entries)
        url = aura_groups.Meta.api.with_query(ids=','.join(aura_ids))
        async with self.manager.web_client.get(
            url, headers={'api-key': self.builder_config.api_key or ''}
        ) as response:
            metadata = await response.json()

        results = chain_dict(
            aura_ids, None, ((i.slug, i) for i in map(WagoResponse.parse_obj, metadata))
        )
        return list(results.values())

    async def get_wago_import_string(self, aura_id: str) -> str:
        async with self.manager.web_client.get(
            import_api.with_query(id=aura_id),
            headers={'api-key': self.builder_config.api_key or ''},
        ) as response:
            return await response.text()

    async def get_remote_auras(
        self, aura_groups: WeakAuras
    ) -> Tuple[Type[WeakAuras], List[RemoteAura]]:
        if not aura_groups.entries:
            return (aura_groups.__class__, [])

        metadata = await self.get_wago_metadata(aura_groups)
        import_strings = await gather((self.get_wago_import_string(m.id) for m in metadata), False)
        return (
            aura_groups.__class__,
            list(zip(aura_groups.entries.values(), metadata, import_strings)),
        )

    def make_addon(self, auras: Sequence[Tuple[Type[WeakAuras], Sequence[RemoteAura]]]) -> None:
        from functools import partial
        from importlib.resources import read_text
        from zipfile import ZipFile, ZipInfo

        from jinja2 import Environment, FunctionLoader

        from . import wa_templates

        jinja_env = Environment(
            trim_blocks=True,
            lstrip_blocks=True,
            loader=FunctionLoader(partial(read_text, wa_templates)),
        )
        aura_dict = chain_dict((WeakAuras, Plateroos), [], auras)

        self.addon_file.parent.mkdir(exist_ok=True)
        with ZipFile(self.addon_file, 'w') as file:

            def write_tpl(filename: str, ctx: Dict[str, Any]) -> None:
                # We're not using a plain string as the first argument to
                # ``writestr`` 'cause then the timestamp is generated dynamically
                # making the build unreproducible
                zip_info = ZipInfo(filename=f'WeakAurasCompanion/{filename}')
                file.writestr(zip_info, jinja_env.get_template(filename).render(ctx))

            write_tpl(
                'data.lua',
                {
                    'weakauras': [
                        (
                            metadata.slug,
                            {
                                'name': metadata.name,
                                'author': metadata.username,
                                'encoded': import_string,
                                'wagoVersion': metadata.version,
                                'wagoSemver': metadata.version_string,
                                'versionNote': metadata.changelog.text,
                            },
                        )
                        for _, metadata, import_string in aura_dict[WeakAuras]
                    ],
                    # Maps internal UIDs of top-level auras to IDs or slugs on Wago
                    'weakaura_uids': [
                        (a.uid, a.url.parts[1])
                        for existing_auras, _, _ in aura_dict[WeakAuras]
                        for a in (
                            next((i for i in existing_auras if not i.parent), existing_auras[0]),
                        )
                    ],
                    # Maps local names to IDs or slugs on Wago
                    'weakaura_ids': [
                        (a.id, a.url.parts[1])
                        for existing_auras, _, _ in aura_dict[WeakAuras]
                        for a in existing_auras
                    ],
                    'plateroos': [
                        (
                            metadata.slug,
                            {
                                'name': metadata.name,
                                'author': metadata.username,
                                'encoded': import_string,
                                'wagoVersion': metadata.version,
                                'wagoSemver': metadata.version_string,
                                'versionNote': metadata.changelog.text,
                            },
                        )
                        for _, metadata, import_string in aura_dict[Plateroos]
                    ],
                    'plater_ids': [
                        (a.id, a.url.parts[1])
                        for existing_auras, _, _ in aura_dict[Plateroos]
                        for a in existing_auras
                    ],
                },
            )
            write_tpl('init.lua', {})
            write_tpl(
                'WeakAurasCompanion.toc',
                {'interface': '11305' if self.manager.config.is_classic else '80300'},
            )

    async def build(self) -> None:
        installed_auras = iit(self.extract_installed_auras())
        remote_auras = await gather(
            [self.get_remote_auras(g) async for g in installed_auras],
            False,
        )
        await t(self.make_addon)(remote_auras)

    def checksum(self) -> str:
        from hashlib import sha256

        return sha256(self.addon_file.read_bytes()).hexdigest()



from functools import lru_cache, partial
from pathlib import Path
import re

import pytest

from instawow.config import Config
from instawow.manager import Manager
from instawow.utils import get_version


def pytest_addoption(parser):
    parser.addoption('--instawow-no-mock', action='store_true')


def should_mock(fn):
    import inspect
    import warnings

    def wrapper(request):
        if request.config.getoption('--instawow-no-mock'):
            warnings.warn('not mocking')
            return None

        args = (request.getfixturevalue(p) for p in inspect.signature(fn).parameters)
        return fn(*args)

    return wrapper


@lru_cache(maxsize=None)
def read_fixture(filename):
    return (Path(__file__).parent / 'fixtures' / filename).read_bytes()


@lru_cache(maxsize=None)
def make_zip(name):
    from io import BytesIO
    from zipfile import ZipFile

    buffer = BytesIO()
    with ZipFile(buffer, 'w') as file:
        file.writestr(f'{name}/{name}.toc', b'')
    return buffer.getvalue()


@pytest.fixture(scope='session')
def temp_dir(tmp_path_factory):
    yield tmp_path_factory.mktemp('temp')


@pytest.fixture(params=['retail', 'classic'])
def partial_config(tmp_path, request, temp_dir):
    addons = tmp_path / 'wow' / 'interface' / 'addons'
    addons.mkdir(parents=True)
    return {'addon_dir': addons, 'temp_dir': temp_dir, 'game_flavour': request.param}


@pytest.fixture
def full_config(tmp_path, partial_config):
    return {**partial_config, 'config_dir': tmp_path / 'config'}


@pytest.fixture
def manager(full_config):
    config = Config(**full_config).write()
    manager = Manager.from_config(config)
    yield manager


@pytest.fixture
def JsonResponse(aresponses):
    return partial(aresponses.Response, headers={'Content-Type': 'application/json'})


@pytest.fixture
@should_mock
def mock_pypi(aresponses, JsonResponse):
    aresponses.add(
        'pypi.org',
        '/pypi/instawow/json',
        'get',
        JsonResponse(body=f'{{"info": {{"version": "{get_version()}"}}}}'),
        repeat=float('inf'),
    )


@pytest.fixture
@should_mock
def mock_master_catalogue(aresponses, JsonResponse):
    aresponses.add(
        'raw.githubusercontent.com',
        aresponses.ANY,
        'get',
        JsonResponse(body=read_fixture('master-catalogue.json')),
        repeat=float('inf'),
    )


@pytest.fixture
@should_mock
def mock_curse(aresponses, JsonResponse, mock_master_catalogue):
    aresponses.add(
        'addons-ecs.forgesvc.net',
        '/api/v2/addon',
        'post',
        JsonResponse(body=read_fixture('curse-addon--all.json')),
        repeat=float('inf'),
    )
    aresponses.add(
        'addons-ecs.forgesvc.net',
        '/api/v2/addon/20338/files',
        'get',
        JsonResponse(body=read_fixture('curse-addon-files.json')),
        repeat=float('inf'),
    )
    aresponses.add(
        'edge.forgecdn.net',
        aresponses.ANY,
        'get',
        aresponses.Response(body=make_zip('Molinari')),
        repeat=float('inf'),
    )


@pytest.fixture
@should_mock
def mock_wowi(aresponses, JsonResponse, mock_master_catalogue):
    aresponses.add(
        'api.mmoui.com',
        '/v3/game/WOW/filelist.json',
        'get',
        JsonResponse(body=read_fixture('wowi-filelist.json')),
        repeat=float('inf'),
    )
    aresponses.add(
        'api.mmoui.com',
        re.compile(r'^/v3/game/WOW/filedetails/'),
        'get',
        JsonResponse(body=read_fixture('wowi-filedetails.json')),
        repeat=float('inf'),
    )
    aresponses.add(
        'cdn.wowinterface.com',
        aresponses.ANY,
        'get',
        aresponses.Response(body=make_zip('Molinari')),
        repeat=float('inf'),
    )


@pytest.fixture
@should_mock
def mock_tukui(aresponses, JsonResponse, mock_master_catalogue):
    aresponses.add(
        'www.tukui.org',
        '/api.php?ui=tukui',
        'get',
        JsonResponse(body=read_fixture('tukui-ui--tukui.json')),
        match_querystring=True,
        repeat=float('inf'),
    )
    aresponses.add(
        'www.tukui.org',
        '/api.php?addon=1',
        'get',
        JsonResponse(body=read_fixture('tukui-addon.json')),
        match_querystring=True,
        repeat=float('inf'),
    )
    aresponses.add(
        'www.tukui.org',
        '/api.php?classic-addon=1',
        'get',
        JsonResponse(body=read_fixture('tukui-classic-addon.json')),
        match_querystring=True,
        repeat=float('inf'),
    )
    aresponses.add('www.tukui.org', '/api.php', 'get', '', repeat=float('inf'))
    aresponses.add(
        'www.tukui.org',
        re.compile(r'^/downloads/tukui'),
        'get',
        aresponses.Response(body=make_zip('Tukui')),
        repeat=float('inf'),
    )
    aresponses.add(
        'www.tukui.org',
        '/addons.php?download=1',
        'get',
        aresponses.Response(body=make_zip('ElvUI_MerathilisUI')),
        match_querystring=True,
        repeat=float('inf'),
    )
    aresponses.add(
        'www.tukui.org',
        '/classic-addons.php?download=1',
        'get',
        aresponses.Response(body=make_zip('Tukui')),
        match_querystring=True,
        repeat=float('inf'),
    )


@pytest.fixture
@should_mock
def mock_github(aresponses, JsonResponse):
    aresponses.add(
        'api.github.com',
        '/repos/AdiAddons/AdiButtonAuras',
        'get',
        JsonResponse(body=read_fixture('github-repo-lib-and-nolib.json')),
        repeat=float('inf'),
    )
    aresponses.add(
        'api.github.com',
        '/repos/AdiAddons/AdiButtonAuras/releases/latest',
        'get',
        JsonResponse(body=read_fixture('github-release-lib-and-nolib.json')),
        repeat=float('inf'),
    )
    aresponses.add(
        'api.github.com',
        '/repos/AdiAddons/AdiButtonAuras/releases/tags/2.1.0',
        'get',
        JsonResponse(body=read_fixture('github-release-lib-and-nolib-older-version.json')),
        repeat=float('inf'),
    )
    aresponses.add(
        'api.github.com',
        '/repos/WeakAuras/WeakAuras2',
        'get',
        JsonResponse(body=read_fixture('github-repo-retail-and-classic.json')),
        repeat=float('inf'),
    )
    aresponses.add(
        'api.github.com',
        '/repos/WeakAuras/WeakAuras2/releases/latest',
        'get',
        JsonResponse(body=read_fixture('github-release-retail-and-classic.json')),
        repeat=float('inf'),
    )
    aresponses.add(
        'api.github.com',
        '/repos/p3lim-wow/Molinari',
        'get',
        JsonResponse(body=read_fixture('github-repo-no-releases.json')),
        repeat=float('inf'),
    )
    aresponses.add(
        'api.github.com',
        '/repos/p3lim-wow/Molinari/releases/latest',
        'get',
        aresponses.Response(body=b'', status=404),
        repeat=float('inf'),
    )
    aresponses.add(
        'api.github.com',
        '/repos/AdiAddons/AdiButtonAuras/releases/tags/2.0.19',
        'get',
        JsonResponse(body=read_fixture('github-release-no-assets.json')),
        repeat=float('inf'),
    )
    aresponses.add(
        'api.github.com',
        '/repos/layday/foo-bar',
        'get',
        aresponses.Response(body=b'', status=404),
        repeat=float('inf'),
    )
    aresponses.add(
        'github.com',
        re.compile(r'^(/[^/]*){2}/releases/download'),
        'get',
        aresponses.Response(body=make_zip('Foo')),
        repeat=float('inf'),
    )


@pytest.fixture
def mock_all(
    mock_pypi,
    mock_curse,
    mock_wowi,
    mock_tukui,
    mock_github,
):
    pass



from __future__ import annotations

import asyncio
from collections import defaultdict, deque
from contextlib import contextmanager
from datetime import datetime, timedelta
from functools import partial, wraps
from itertools import chain, repeat, takewhile
from pathlib import Path, PurePath
import posixpath
import re
from shutil import move as _move
from tempfile import mkdtemp
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    DefaultDict,
    Dict,
    Generic,
    Hashable,
    Iterable,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

from typing_extensions import Literal

if TYPE_CHECKING:
    from prompt_toolkit.shortcuts import ProgressBar

    _H = TypeVar('_H', bound=Hashable)
    _AnySet = TypeVar('_AnySet', bound=AbstractSet)


_V = TypeVar('_V')


class _TocEntry(NamedTuple):
    key: str
    value: str

    def __bool__(self) -> bool:
        return bool(self.value)


class TocReader:
    """Extracts key–value pairs from TOC files."""

    default = ''

    def __init__(self, contents: str) -> None:
        possible_entries = (
            map(str.strip, e.lstrip('#').partition(':')[::2])
            for e in contents.splitlines()
            if e.startswith('##')
        )
        self.entries = {k: v for k, v in possible_entries if k}

    def __getitem__(self, key: Union[str, Tuple[str, ...]]) -> _TocEntry:
        if isinstance(key, tuple):
            try:
                return next(filter(None, (self[k] for k in key)))
            except StopIteration:
                key = key[0]
        return _TocEntry(key, self.entries.get(key, self.default))

    @classmethod
    def from_path(cls, path: Path) -> TocReader:
        return cls(path.read_text(encoding='utf-8-sig', errors='replace'))

    @classmethod
    def from_parent_folder(cls, path: Path) -> TocReader:
        return cls.from_path(path / f'{path.name}.toc')


class cached_property(Generic[_V]):
    def __init__(self, f: Callable[[Any], _V]) -> None:
        self.f = f

    @overload
    def __get__(self, o: None, t: Optional[type] = None) -> cached_property[_V]:
        ...

    @overload
    def __get__(self, o: Any, t: Optional[type] = None) -> _V:
        ...

    def __get__(self, o: Any, t: Optional[type] = None) -> Union[cached_property[_V], _V]:
        if o is None:
            return self
        else:
            o.__dict__[self.f.__name__] = v = self.f(o)
            return v


def bucketise(iterable: Iterable[_V], key: Callable[[_V], _H]) -> DefaultDict[_H, List[_V]]:
    "Place the elements of an iterable in a bucket according to ``key``."
    bucket: Any = defaultdict(list)
    for value in iterable:
        bucket[key(value)].append(value)
    return bucket


def chain_dict(
    keys: Iterable[_H], default: Any, *overrides: Iterable[Tuple[_H, Any]]
) -> Dict[_H, Any]:
    "Construct a dictionary from a series of iterables with overlapping keys."
    return dict(chain(zip(keys, repeat(default)), *overrides))


def uniq(it: Iterable[_H]) -> List[_H]:
    "Deduplicate hashable items in an iterable maintaining insertion order."
    return list(dict.fromkeys(it))


def merge_intersecting_sets(it: Iterable[_AnySet]) -> Iterable[_AnySet]:
    "Recursively merge intersecting sets in a collection."
    many_sets = deque(it)
    while True:
        try:
            this_set = many_sets.popleft()
        except IndexError:
            return

        while True:
            for other_set in many_sets:
                if this_set & other_set:
                    # The in-place operator would mutate unfrozen sets
                    # in the original collection
                    this_set = this_set | other_set  # type: ignore
                    many_sets.remove(other_set)
                    break
            else:
                break
        yield this_set


@overload
async def gather(
    it: Iterable[Awaitable[_V]], return_exceptions: Literal[True] = ...
) -> List[Union[BaseException, _V]]:
    ...


@overload
async def gather(it: Iterable[Awaitable[_V]], return_exceptions: Literal[False] = ...) -> List[_V]:
    ...


async def gather(
    it: Iterable[Awaitable[_V]], return_exceptions: bool = True
) -> Union[List[Union[BaseException, _V]], List[_V]]:
    return await asyncio.gather(*it, return_exceptions=return_exceptions)


def run_in_thread(fn: Callable[..., _V]) -> Callable[..., Awaitable[_V]]:
    @wraps(fn)
    def wrapper(*args: Any, **kwargs: Any):
        loop = asyncio.get_running_loop()
        return loop.run_in_executor(None, lambda: fn(*args, **kwargs))

    return wrapper


class _StopIteration(Exception):
    pass


def _iit_next(it: Iterator[_V]) -> _V:
    try:
        return next(it)
    except StopIteration:
        raise _StopIteration


# From Starlette: https://github.com/encode/starlette/blob/78f7095/starlette/concurrency.py
async def iter_in_thread(it: Iterator[_V]) -> AsyncIterator[_V]:
    loop = asyncio.get_running_loop()
    while True:
        try:
            yield await loop.run_in_executor(None, _iit_next, it)
        except _StopIteration:
            return


@contextmanager
def copy_resources(*packages: str) -> Iterator[Path]:
    """Copy package resources to a temporary directory on disk.

    Alembic cannot construct a migration environment from memory.
    This is a genericised bodge to make migrations work in a frozen instawow.
    """
    from importlib.resources import contents, is_resource, read_binary
    from tempfile import TemporaryDirectory

    with TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)

        for package in packages:
            for resource in filter(partial(is_resource, package), contents(package)):
                parent_dir = tmp_path.joinpath(*package.split('.'))
                if not parent_dir.is_dir():
                    parent_dir.mkdir(parents=True)

                filename = parent_dir / resource
                # PyOxidizer does not expose Python source files to `importlib`
                # (see https://github.com/indygreg/PyOxidizer/issues/237).
                # The filenames of Python files have a secondary ".0" extension
                # so that PyOx will regard them as binary resources.
                # The extension is then removed when they're copied into
                # the temporary folder
                if filename.suffix == '.0':
                    filename = filename.with_suffix('')
                filename.write_bytes(read_binary(package, resource))
        yield tmp_path


def slugify(text: str) -> str:
    "Convert an add-on name into a lower-alphanumeric slug."
    return '-'.join(re.sub(r'[^0-9a-z]', ' ', text.casefold()).split())


def tabulate(rows: Sequence[Sequence[Any]], *, max_col_width: int = 60) -> str:
    "Produce an ASCII table from equal-length elements in a sequence."
    from textwrap import fill

    def apply_max_col_width(value: Sequence[Any]):
        return fill(str(value), width=max_col_width, max_lines=1)

    def calc_resultant_col_widths(rows: Sequence[Sequence[str]]):
        cols = zip(*rows)
        return [max(map(len, c)) for c in cols]

    rows = [tuple(map(apply_max_col_width, r)) for r in rows]
    head, *tail = rows

    base_template = '  '.join(f'{{{{{{0}}{w}}}}}' for w in calc_resultant_col_widths(rows))
    row_template = base_template.format(':<')
    table = '\n'.join(
        (
            base_template.format(':^').format(*head),
            base_template.format(f'0:-<').format(''),
            *(row_template.format(*r) for r in tail),
        )
    )
    return table


def make_progress_bar(**kwargs: Any) -> ProgressBar:
    "A ``ProgressBar`` with download progress expressed in megabytes."
    import signal

    from prompt_toolkit.formatted_text import HTML
    from prompt_toolkit.shortcuts.progress_bar import ProgressBar, formatters
    from prompt_toolkit.utils import Event

    # There is a race condition in the bar's shutdown logic
    # where the bar-drawing thread is left to run indefinitely
    # if the prompt_toolkit app is not (yet) running
    # but is scheduled to run when ``__exit__``ing the bar.
    # If an exception occurs early in the execution cycle or if execution
    # finishes before prompt_toolkit is able to crank the app,
    # instawow will hang.  This is my ham-fisted attempt to work
    # around all that by signalling to the daemon thread to kill the app.

    _SIGWINCH = getattr(signal, "SIGWINCH", None)

    class PatchedProgressBar(ProgressBar):
        def __exit__(self, *args: Any):
            if self._has_sigwinch:
                self._loop.remove_signal_handler(_SIGWINCH)
                signal.signal(_SIGWINCH, self._previous_winch_handler)

            if self._thread is not None:

                def attempt_exit(sender: Any):
                    sender.is_running and sender.exit()

                # Signal to ``_auto_refresh_context`` that it should exit the app
                self.app.on_invalidate = Event(self.app, attempt_exit)
                self._thread.join()

            self._app_loop.close()

    class DownloadProgress(formatters.Progress):
        template = formatters.Progress.template + 'MB'

        def format(self, progress_bar: ProgressBar, progress: Any, width: int):
            def format_pct(value: int) -> str:
                return f'{value / 2 ** 20:.1f}'

            return HTML(self.template).format(
                current=format_pct(progress.items_completed),
                total=format_pct(progress.total) if progress.total else '?',
            )

    f = [
        formatters.Label(),
        formatters.Text(' '),
        formatters.Percentage(),
        formatters.Text(' '),
        formatters.Bar(),
        formatters.Text(' '),
        DownloadProgress(),
        formatters.Text(' '),
        formatters.Text('eta [', style='class:time-left'),
        formatters.TimeLeft(),
        formatters.Text(']', style='class:time-left'),
        formatters.Text(' '),
    ]
    progress_bar = PatchedProgressBar(formatters=f, **kwargs)
    return progress_bar


def move(src: PurePath, dst: PurePath) -> PurePath:
    # See https://bugs.python.org/issue32689
    return _move(str(src), dst)


def trash(paths: Sequence[PurePath], *, dst: PurePath, missing_ok: bool = False) -> None:
    parent_folder = Path(mkdtemp(dir=dst, prefix=f'deleted-{paths[0].name}-'))
    for path in paths:
        try:
            move(path, dst=parent_folder)
        except (FileNotFoundError if missing_ok else ()):
            pass


def shasum(*values: str) -> str:
    "Base-16-encode a string using SHA-256 truncated to 32 characters."
    from hashlib import sha256

    return sha256(''.join(values).encode()).hexdigest()[:32]


def is_not_stale(path: Path, ttl: int, unit: str = 'seconds') -> bool:
    "Check if a file is older than ``ttl``."
    mtime = path.exists() and path.stat().st_mtime
    return mtime > 0 and (
        (datetime.now() - datetime.fromtimestamp(cast(float, mtime))) < timedelta(**{unit: ttl})
    )


def get_version() -> str:
    "Get the installed version of instawow."
    try:
        from ._version import __version__
    except ImportError:
        return 'dev'
    else:
        return __version__


def is_outdated() -> Tuple[bool, str]:
    """Check on PyPI to see if the installed copy of instawow is outdated.

    The response is cached for 24 hours.
    """
    __version__ = get_version()
    if 'dev' in __version__:
        return (False, '')

    from aiohttp.client import ClientError

    from .config import Config
    from .manager import init_web_client

    def parse_version(version: str) -> Tuple[int, ...]:
        version_numbers = version.split('.')[:3]
        int_only_version_numbers = (
            int(''.join(takewhile('0123456789'.__contains__, e))) for e in version_numbers
        )
        return tuple(int_only_version_numbers)

    dummy_config = Config.get_dummy_config()
    if not dummy_config.auto_update_check:
        return (False, '')

    cache_file = dummy_config.temp_dir / '.pypi_version'
    if is_not_stale(cache_file, 1, 'days'):
        version = cache_file.read_text(encoding='utf-8')
    else:

        async def get_metadata():
            api_url = 'https://pypi.org/pypi/instawow/json'
            async with init_web_client() as web_client, web_client.get(api_url) as response:
                return await response.json()

        loop = asyncio.new_event_loop()
        try:
            version = loop.run_until_complete(get_metadata())['info']['version']
        except ClientError:
            version = __version__
        else:
            cache_file.write_text(version, encoding='utf-8')
        finally:
            loop.close()

    return (parse_version(version) > parse_version(__version__), version)


def find_zip_base_dirs(names: Sequence[str]) -> Set[str]:
    "Find top-level folders in a list of ZIP member paths."
    return {n for n in (posixpath.dirname(n) for n in names) if n and posixpath.sep not in n}


def make_zip_member_filter(base_dirs: Set[str]) -> Callable[[str], bool]:
    """Filter out items which are not sub-paths of top-level folders for the
    purpose of extracting ZIP files.
    """

    def is_subpath(name: str):
        head, sep, _ = name.partition(posixpath.sep)
        return cast(bool, sep) and head in base_dirs

    return is_subpath



from __future__ import annotations

from datetime import datetime
from typing import Any, ClassVar, List, Type, cast

import pydantic
from sqlalchemy import (
    Column,
    DateTime,
    ForeignKeyConstraint,
    Integer,
    MetaData,
    String,
    UniqueConstraint,
    and_,
    exc,
    func,
    inspect,
)
from sqlalchemy.ext.declarative import DeclarativeMeta, as_declarative
from sqlalchemy.orm import object_session, relationship


class _BaseCoercer(pydantic.BaseModel):
    """The coercer is used inside the declarative constructor to type-cast
    values _prior to_ inserts.  SQLAlchemy delegates this function to
    the DB API -- see https://stackoverflow.com/a/8980982.  We need this to
    happen in advance to be able to compare values with their in-database
    counterparts.  It also saves us the trouble of having to manually
    parse dates and the like.
    """

    class Config:
        extra = pydantic.Extra.allow
        max_anystr_length = 2 ** 32
        orm_mode = True


class _BaseTableMeta(DeclarativeMeta):
    def __init__(cls, *args: Any) -> None:
        super().__init__(*args)
        try:
            inspector = inspect(cls)
        except exc.NoInspectionAvailable:
            pass
        else:
            cls.Coercer = pydantic.create_model(
                f'{cls.__name__}Coercer',
                __base__=_BaseCoercer,
                **{
                    c.name: (c.type.python_type, ...)
                    for c in inspector.columns
                    if not c.foreign_keys and not c.name.startswith('_') and not c.server_default
                },
            )


@as_declarative(constructor=None, metaclass=_BaseTableMeta)
class _BaseTable:
    Coercer: ClassVar[Type[_BaseCoercer]]
    metadata: ClassVar[MetaData]

    def __init__(self, **kwargs: Any) -> None:
        intermediate_obj = self.Coercer(**kwargs)
        for k, v in intermediate_obj:
            setattr(self, k, v)


ModelBase = _BaseTable


class Pkg(_BaseTable):
    __tablename__ = 'pkg'

    source = Column(String, primary_key=True)
    id = Column(String, primary_key=True)
    slug = Column(String, nullable=False)
    name = Column(String, nullable=False)
    description = Column(String, nullable=False)
    url = Column(String, nullable=False)
    download_url = Column(String, nullable=False)
    date_published = Column(DateTime, nullable=False)
    version = Column(String, nullable=False)
    folders = cast(
        'List[PkgFolder]', relationship('PkgFolder', cascade='all, delete-orphan', backref='pkg')
    )
    options = cast(
        'PkgOptions', relationship('PkgOptions', cascade='all, delete-orphan', uselist=False)
    )
    deps = cast(
        'List[PkgDep]', relationship('PkgDep', cascade='all, delete-orphan', backref='pkg')
    )

    @property
    def logged_versions(self) -> List[PkgVersionLog]:
        session: Any = object_session(self)
        return (
            (
                session.query(PkgVersionLog)
                .filter(
                    and_(PkgVersionLog.pkg_source == self.source, PkgVersionLog.pkg_id == self.id)
                )
                .order_by(PkgVersionLog.install_time.desc())
                .all()
            )
            if session
            else []
        )


class PkgFolder(_BaseTable):
    __tablename__ = 'pkg_folder'
    __table_args__ = (ForeignKeyConstraint(['pkg_source', 'pkg_id'], ['pkg.source', 'pkg.id']),)

    name = Column(String, primary_key=True)
    pkg_source = Column(String, nullable=False)
    pkg_id = Column(String, nullable=False)


class PkgOptions(_BaseTable):
    __tablename__ = 'pkg_options'
    __table_args__ = (ForeignKeyConstraint(['pkg_source', 'pkg_id'], ['pkg.source', 'pkg.id']),)

    strategy = Column(String, nullable=False)
    pkg_source = Column(String, primary_key=True)
    pkg_id = Column(String, primary_key=True)


class PkgDep(_BaseTable):
    __tablename__ = 'pkg_dep'
    __table_args__ = (
        ForeignKeyConstraint(['pkg_source', 'pkg_id'], ['pkg.source', 'pkg.id']),
        UniqueConstraint('id', 'pkg_source', 'pkg_id', name='uq_id_per_foreign_key_constr'),
    )

    _id = Column(Integer, primary_key=True)
    id = Column(String, nullable=False)
    pkg_source = Column(String, nullable=False)
    pkg_id = Column(String, nullable=False)


class PkgVersionLog(_BaseTable):
    __tablename__ = 'pkg_version_log'

    version = Column(String, nullable=False, primary_key=True)
    install_time = Column(DateTime, nullable=False, server_default=func.now())
    pkg_source = Column(String, nullable=False, primary_key=True)
    pkg_id = Column(String, nullable=False, primary_key=True)


class _PkgFolder(_BaseCoercer):
    name: str


class _PkgOptions(_BaseCoercer):
    strategy: str


class _PkgDep(_BaseCoercer):
    id: str


class _PkgVersion(_BaseCoercer):
    version: str
    install_time: datetime


class PkgModel(_BaseCoercer):
    source: str
    id: str
    slug: str
    name: str
    description: str
    url: str
    download_url: str
    date_published: datetime
    version: str
    folders: List[_PkgFolder]
    options: _PkgOptions
    deps: List[_PkgDep]
    logged_versions: List[_PkgVersion]


class MultiPkgModel(pydantic.BaseModel):
    __root__: List[PkgModel]


def should_migrate(engine: Any, version: str) -> bool:
    """Check if the database version is the same as `version`;
    if not a migration would be required.
    """
    with engine.begin() as conn:
        try:
            current = conn.execute(
                'SELECT version_num FROM alembic_version WHERE version_num = (?)',
                version,
            ).scalar()
        except exc.OperationalError:
            return True
        else:
            return not current


def is_pkg(value: Any) -> bool:
    return isinstance(value, Pkg)



import enum
from functools import partial
import json
from types import SimpleNamespace
from unittest.mock import patch

from click.testing import CliRunner
import pytest

from instawow.cli import main
from instawow.config import Config
from instawow.manager import CliManager


class Flavour(enum.IntEnum):
    retail = enum.auto()
    classic = enum.auto()


@pytest.fixture(autouse=True)
def mock(mock_all):
    pass


@pytest.fixture(params=['retail', 'classic'])
def full_config(tmp_path_factory, request, temp_dir):
    name = f'{request.node.name[: request.node.name.index("[")]}_{request.param}'
    parametrized_tmp_path = tmp_path_factory.getbasetemp() / name
    addons = parametrized_tmp_path / 'addons'
    addons.mkdir(parents=True, exist_ok=True)

    return {
        '_parametrized_tmp_path': parametrized_tmp_path,
        'config_dir': parametrized_tmp_path / 'config',
        'addon_dir': addons,
        'temp_dir': temp_dir,
        'game_flavour': request.param,
    }


@pytest.fixture
def manager(event_loop, full_config):
    config = Config(**full_config).write()
    manager = CliManager.from_config(config)
    manager.run = event_loop.run_until_complete
    yield manager


@pytest.fixture
def run(request, manager):
    param = getattr(request, 'param', None)
    if param is Flavour.retail and manager.config.is_classic:
        pytest.skip('test is for retail only')
    elif param is Flavour.classic and manager.config.is_retail:
        pytest.skip('test is for classic only')

    yield partial(CliRunner().invoke, main, obj=SimpleNamespace(m=manager))


@pytest.fixture
def molinari_and_run(run):
    run('install curse:molinari')
    yield run


@pytest.mark.parametrize(
    'inp, cmp',
    [
        ('install curse:molinari', lambda v: v.startswith('✓ curse:molinari\n  installed')),
        ('install curse:molinari', '✗ curse:molinari\n  package already installed\n'.__eq__),
        ('update curse:molinari', '✗ curse:molinari\n  package is up to date\n'.__eq__),
        ('remove curse:molinari', '✓ curse:molinari\n  removed\n'.__eq__),
        ('update curse:molinari', '✗ curse:molinari\n  package is not installed\n'.__eq__),
        ('remove curse:molinari', '✗ curse:molinari\n  package is not installed\n'.__eq__),
    ],
)
def test_valid_curse_pkg_lifecycle(run, inp, cmp):
    assert cmp(run(inp).output)


@pytest.mark.parametrize(
    'inp, cmp, run',
    [
        ('install tukui:1', lambda v: v.startswith('✓ tukui:1\n  installed'), None),
        ('install tukui:1', '✗ tukui:1\n  package already installed\n'.__eq__, None),
        (
            'update tukui:1',
            '✗ tukui:1-merathilisui\n  package is up to date\n'.__eq__,
            Flavour.retail,
        ),
        (
            'update tukui:1',
            '✗ tukui:1-tukui\n  package is up to date\n'.__eq__,
            Flavour.classic,
        ),
        ('remove tukui:1', '✓ tukui:1\n  removed\n'.__eq__, None),
        ('update tukui:1', '✗ tukui:1\n  package is not installed\n'.__eq__, None),
        ('remove tukui:1', '✗ tukui:1\n  package is not installed\n'.__eq__, None),
        (
            'install tukui:tukui',
            lambda v: v.startswith('✓ tukui:tukui\n  installed'),
            Flavour.retail,
        ),
        (
            'install tukui:tukui',
            '✗ tukui:tukui\n  package already installed\n'.__eq__,
            Flavour.retail,
        ),
        ('update tukui:tukui', '✗ tukui:tukui\n  package is up to date\n'.__eq__, Flavour.retail),
        ('remove tukui:tukui', '✓ tukui:tukui\n  removed\n'.__eq__, Flavour.retail),
        (
            'install tukui:tukui',
            '✗ tukui:tukui\n  package does not exist\n'.__eq__,
            Flavour.classic,
        ),
        ('update tukui:tukui', '✗ tukui:tukui\n  package is not installed\n'.__eq__, None),
        ('remove tukui:tukui', '✗ tukui:tukui\n  package is not installed\n'.__eq__, None),
    ],
    indirect=('run',),
)
def test_valid_tukui_pkg_lifecycle(run, inp, cmp):
    assert cmp(run(inp).output)


@pytest.mark.parametrize(
    'inp, cmp',
    [
        (
            'install wowi:13188-molinari',
            lambda v: v.startswith('✓ wowi:13188-molinari\n  installed'),
        ),
        (
            'install wowi:13188-molinari',
            '✗ wowi:13188-molinari\n  package already installed\n'.__eq__,
        ),
        ('update wowi:13188-molinari', '✗ wowi:13188-molinari\n  package is up to date\n'.__eq__),
        ('remove wowi:13188-molinari', '✓ wowi:13188-molinari\n  removed\n'.__eq__),
        (
            'update wowi:13188-molinari',
            '✗ wowi:13188-molinari\n  package is not installed\n'.__eq__,
        ),
        (
            'remove wowi:13188-molinari',
            '✗ wowi:13188-molinari\n  package is not installed\n'.__eq__,
        ),
    ],
)
def test_valid_wowi_pkg_lifecycle(run, inp, cmp):
    assert cmp(run(inp).output)


@pytest.mark.parametrize(
    'inp, cmp, run',
    [
        (
            'install curse:molinari',
            lambda v: v.startswith('✓ curse:molinari\n  installed'),
            Flavour.retail,
        ),
        (
            'install wowi:13188-molinari',
            "✗ wowi:13188-molinari\n"
            "  package folders conflict with installed package curse:molinari\n".__eq__,
            Flavour.retail,
        ),
        ('remove curse:molinari', '✓ curse:molinari\n  removed\n'.__eq__, Flavour.retail),
    ],
    indirect=('run',),
)
def test_folder_conflict_lifecycle(run, inp, cmp):
    assert cmp(run(inp).output)


def test_preexisting_folder_conflict_on_install(manager, run):
    manager.config.addon_dir.joinpath('Molinari').mkdir()
    assert (
        run('install curse:molinari').output
        == '''\
✗ curse:molinari
  package folders conflict with 'Molinari'
'''
    )
    assert run('install -o curse:molinari').output.startswith('✓ curse:molinari\n  installed')


@pytest.mark.parametrize(
    'inp, output',
    [
        ('install curse:gargantuan-wigs', '✗ curse:gargantuan-wigs\n  package does not exist\n'),
        ('update curse:gargantuan-wigs', '✗ curse:gargantuan-wigs\n  package is not installed\n'),
        ('remove curse:gargantuan-wigs', '✗ curse:gargantuan-wigs\n  package is not installed\n'),
    ],
)
def test_invalid_addon_name_lifecycle(run, inp, output):
    assert run(inp).output == output


@pytest.mark.parametrize(
    'inp, output',
    [
        ('install foo:bar', '✗ foo:bar\n  package source is invalid\n'),
        ('update foo:bar', '✗ foo:bar\n  package is not installed\n'),
        ('remove foo:bar', '✗ foo:bar\n  package is not installed\n'),
    ],
)
def test_invalid_source_lifecycle(run, inp, output):
    assert run(inp).output == output


@pytest.mark.parametrize(
    'inp, cmp',
    [
        ('install curse:molinari', lambda v: v.startswith('✓ curse:molinari\n  installed')),
        ('install curse:20338', '✗ curse:20338\n  package already installed\n'.__eq__),
        (
            'install https://www.wowace.com/projects/molinari',
            '✗ curse:molinari\n  package already installed\n'.__eq__,
        ),
        (
            'install https://www.curseforge.com/wow/addons/molinari',
            '✗ curse:molinari\n  package already installed\n'.__eq__,
        ),
        (
            'install https://www.curseforge.com/wow/addons/molinari/download',
            '✗ curse:molinari\n  package already installed\n'.__eq__,
        ),
        ('remove curse:molinari', '✓ curse:molinari\n  removed\n'.__eq__),
    ],
)
def test_install_with_curse_alias(run, inp, cmp):
    assert cmp(run(inp).output)


@pytest.mark.parametrize(
    'inp, cmp, run',
    [
        ('install tukui:-1', lambda v: v.startswith('✓ tukui:-1\n  installed'), Flavour.retail),
        ('install tukui:-1', '✗ tukui:-1\n  package does not exist\n'.__eq__, Flavour.classic),
        (
            'install tukui:tukui',
            '✗ tukui:tukui\n  package already installed\n'.__eq__,
            Flavour.retail,
        ),
        (
            'install tukui:tukui',
            '✗ tukui:tukui\n  package does not exist\n'.__eq__,
            Flavour.classic,
        ),
        (
            'install https://www.tukui.org/download.php?ui=tukui',
            '✗ tukui:tukui\n  package already installed\n'.__eq__,
            Flavour.retail,
        ),
        (
            'install https://www.tukui.org/download.php?ui=tukui',
            '✗ tukui:tukui\n  package does not exist\n'.__eq__,
            Flavour.classic,
        ),
        (
            'install https://www.tukui.org/addons.php?id=1',
            lambda v: v.startswith('✓ tukui:1\n  installed'),
            None,
        ),
        ('install tukui:1', '✗ tukui:1\n  package already installed\n'.__eq__, None),
        ('remove tukui:1', '✓ tukui:1\n  removed\n'.__eq__, None),
    ],
    indirect=('run',),
)
def test_install_with_tukui_alias(run, inp, cmp):
    assert cmp(run(inp).output)


@pytest.mark.parametrize(
    'inp, cmp',
    [
        ('install wowi:13188', lambda v: v.startswith('✓ wowi:13188\n  installed')),
        (
            'install wowi:13188-molinari',
            '✗ wowi:13188-molinari\n  package already installed\n'.__eq__,
        ),
        (
            'install https://www.wowinterface.com/downloads/landing.php?fileid=13188',
            '✗ wowi:13188\n  package already installed\n'.__eq__,
        ),
        (
            'install https://www.wowinterface.com/downloads/fileinfo.php?id=13188',
            '✗ wowi:13188\n  package already installed\n'.__eq__,
        ),
        (
            'install https://www.wowinterface.com/downloads/download13188-Molinari',
            '✗ wowi:13188\n  package already installed\n'.__eq__,
        ),
        (
            'install https://www.wowinterface.com/downloads/info13188-Molinari.html',
            '✗ wowi:13188\n  package already installed\n'.__eq__,
        ),
        ('remove wowi:13188', '✓ wowi:13188\n  removed\n'.__eq__),
    ],
)
def test_install_with_wowi_alias(run, inp, cmp):
    assert cmp(run(inp).output)


@pytest.mark.parametrize(
    'inp, cmp',
    [
        (
            'install github:AdiAddons/AdiButtonAuras',
            lambda v: v.startswith('✓ github:AdiAddons/AdiButtonAuras\n  installed'),
        ),
        (
            'install github:adiaddons/adibuttonauras',
            '✗ github:adiaddons/adibuttonauras\n  package already installed\n'.__eq__,
        ),
        (
            'install https://github.com/AdiAddons/AdiButtonAuras',
            '✗ github:AdiAddons/AdiButtonAuras\n  package already installed\n'.__eq__,
        ),
        (
            'install https://github.com/AdiAddons/AdiButtonAuras/releases',
            '✗ github:AdiAddons/AdiButtonAuras\n  package already installed\n'.__eq__,
        ),
        (
            'remove github:adiaddons/adibuttonauras',
            '✓ github:adiaddons/adibuttonauras\n  removed\n'.__eq__,
        ),
    ],
)
def test_install_with_github_alias(run, inp, cmp):
    assert cmp(run(inp).output)


@pytest.mark.parametrize(
    'inp, output, run',
    [
        ('install curse:molinari', '✓ curse:molinari\n  installed 80300.66-Release\n', None),
        (
            'install --version foo curse:molinari',
            '✗ curse:molinari\n  package already installed\n',
            None,
        ),
        ('update curse:molinari', '✗ curse:molinari\n  package is up to date\n', None),
        ('remove curse:molinari', '✓ curse:molinari\n  removed\n', None),
        (
            'install --version foo curse:molinari',
            "✗ curse:molinari\n  no files compatible with retail using 'version' strategy\n",
            Flavour.retail,
        ),
        (
            'install --version foo curse:molinari',
            "✗ curse:molinari\n  no files compatible with classic using 'version' strategy\n",
            Flavour.classic,
        ),
        (
            'install --version 80000.57-Release curse:molinari',
            '✓ curse:molinari\n  installed 80000.57-Release\n',
            None,
        ),
        ('update curse:molinari', '✗ curse:molinari\n  package is up to date\n', None),
        ('remove curse:molinari', '✓ curse:molinari\n  removed\n', None),
    ],
    indirect=('run',),
)
def test_version_strategy_lifecycle(run, inp, output):
    assert run(inp).output == output


def test_missing_dir_on_remove(manager, molinari_and_run):
    addon_dir = manager.config.addon_dir
    addon_dir.joinpath('Molinari').rename(addon_dir.joinpath('NotMolinari'))
    assert molinari_and_run('remove curse:molinari').output == '✓ curse:molinari\n  removed\n'


@pytest.mark.parametrize(
    'command, exit_code',
    [
        ('list mol', 0),
        ('info foo', 0),
        ('reveal mol', 0),
        ('reveal foo', 1),
    ],
)
@patch('webbrowser.open', lambda v: ...)
def test_exit_codes_with_substr_match(molinari_and_run, command, exit_code):
    assert molinari_and_run(command).exit_code == exit_code


def test_substr_list_match(molinari_and_run):
    assert molinari_and_run('list mol').output == 'curse:molinari\n'
    assert molinari_and_run('list foo').output == ''
    assert molinari_and_run('list -f detailed mol').output.startswith('curse:molinari')
    (molinari,) = json.loads(molinari_and_run('list -f json mol').output)
    assert (molinari['source'], molinari['slug']) == ('curse', 'molinari')


def test_json_export_and_import(molinari_and_run, manager):
    export_json = manager.config._parametrized_tmp_path / 'export.json'
    export_json.write_text(molinari_and_run('list -f json').output, encoding='utf-8')
    assert (
        molinari_and_run(f'install --import "{export_json}"').output
        == '✗ curse:molinari\n  package already installed\n'
    )



from filecmp import dircmp
from pathlib import Path
from shutil import move
import sys

import pytest

from instawow.config import Config


def test_env_vars_have_prio(full_config, monkeypatch):
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', '/foo')
    monkeypatch.setenv('INSTAWOW_GAME_FLAVOUR', 'classic')

    config = Config(**full_config)
    assert config.config_dir == Path('/foo').resolve()
    assert config.game_flavour == 'classic'


def test_config_dir_is_populated(full_config):
    config = Config(**full_config).write()
    assert {i.name for i in config.profile_dir.iterdir()} == {'config.json', 'logs', 'plugins'}


def test_reading_missing_config_from_env_raises(full_config, monkeypatch):
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', str(full_config['config_dir']))
    with pytest.raises(FileNotFoundError):
        Config.read('__default__')


@pytest.mark.skipif(sys.platform == 'win32', reason='no ~ expansion on Windows')
@pytest.mark.parametrize('dir_', ['config_dir', 'addon_dir', 'temp_dir'])
def test_invalid_user_expansion_raises(full_config, dir_):
    with pytest.raises(ValueError):
        Config(**{**full_config, dir_: '~foo'})


def test_missing_addon_dir_raises(full_config):
    with pytest.raises(ValueError):
        Config(**{**full_config, 'addon_dir': 'foo'})


@pytest.mark.skipif(sys.platform == 'win32', reason='path handling')
def test_default_config_dir_is_platform_appropriate(partial_config, monkeypatch):
    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'linux')
        config_dir = Config(**partial_config).config_dir
        assert config_dir == Path.home() / '.config/instawow'

        patcher.setenv('XDG_CONFIG_HOME', '/foo')
        config_dir = Config(**partial_config).config_dir
        assert config_dir == Path('/foo/instawow')

    with monkeypatch.context() as patcher:
        patcher.setattr(sys, 'platform', 'darwin')
        config_dir = Config(**partial_config).config_dir
        assert config_dir == Path.home() / 'Library/Application Support/instawow'


@pytest.mark.skipif(sys.platform != 'win32', reason='path unhandling')
def test_default_config_dir_is_win32_appropriate(partial_config, monkeypatch):
    assert Config(**partial_config).config_dir == Path.home() / 'AppData/Roaming/instawow'
    monkeypatch.delenv('APPDATA')
    assert Config(**partial_config).config_dir == Path.home() / 'instawow'


def test_legacy_profile_migration_goes_swimmingly(full_config, monkeypatch):
    legacy_config = Config(**full_config).write()
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', str(full_config['config_dir']))

    comparison = dircmp(legacy_config.config_dir, legacy_config.profile_dir)
    profile_dirs = sorted(i.name for i in legacy_config.profile_dir.iterdir())
    assert comparison.common == []
    assert comparison.left_list == ['profiles']
    assert sorted(comparison.right_list) == profile_dirs

    for name in profile_dirs:
        move(str(legacy_config.profile_dir / name), legacy_config.config_dir / name)
    legacy_config.profile_dir.rmdir()
    Config.get_dummy_config(config_dir=full_config['config_dir']).read('__default__')

    comparison = dircmp(legacy_config.config_dir, legacy_config.profile_dir)
    assert comparison.common == []
    assert comparison.left_list == ['profiles']
    assert sorted(comparison.right_list) == profile_dirs



from __future__ import annotations

import enum
from functools import partial
from itertools import chain
from pathlib import Path
from textwrap import dedent, fill
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    FrozenSet,
    Generator,
    Generic,
    Iterable,
    List,
    Optional as O,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

import click

from . import exceptions as E, models
from .config import Config, setup_logging
from .resolvers import Defn, Strategies
from .utils import TocReader, cached_property, get_version, is_outdated, tabulate, uniq

if TYPE_CHECKING:
    from .manager import CliManager

    _R = TypeVar('_R')

_EnumT = TypeVar('_EnumT', bound=enum.Enum)


class Report:
    _success_symbol = click.style('✓', fg='green')
    _failure_symbol = click.style('✗', fg='red')
    _warning_symbol = click.style('!', fg='blue')

    def __init__(
        self,
        results: Iterable[Tuple[Defn, E.ManagerResult]],
        filter_fn: Callable[[E.ManagerResult], bool] = lambda _: True,
    ):
        self.results = list(results)
        self.filter_fn = filter_fn

    @property
    def exit_code(self) -> int:
        return any(
            r
            for _, r in self.results
            if isinstance(r, (E.ManagerError, E.InternalError)) and self.filter_fn(r)
        )

    def __str__(self) -> str:
        def adorn_result(result: E.ManagerResult):
            if isinstance(result, E.InternalError):
                return self._warning_symbol
            elif isinstance(result, E.ManagerError):
                return self._failure_symbol
            return self._success_symbol

        return '\n'.join(
            f'{adorn_result(r)} {click.style(str(a), bold=True)}\n'
            + fill(r.message, initial_indent=' ' * 2, subsequent_indent=' ' * 4)
            for a, r in self.results
            if self.filter_fn(r)
        )

    def generate(self):
        manager: CliManager = click.get_current_context().obj.m
        if manager.config.auto_update_check:
            outdated, new_version = is_outdated()
            if outdated:
                click.echo(f'{self._warning_symbol} instawow-{new_version} is available')

        report = str(self)
        if report:
            click.echo(report)

    def generate_and_exit(self):
        self.generate()
        ctx = click.get_current_context()
        ctx.exit(self.exit_code)


def _override_loop_policy() -> None:
    # The proactor event loop which became the default on Windows in 3.8 is
    # tripping up aiohttp<4.  See https://github.com/aio-libs/aiohttp/issues/4324
    import asyncio

    policy = getattr(asyncio, 'WindowsSelectorEventLoopPolicy', None)
    if policy:
        asyncio.set_event_loop_policy(policy())


class ManagerWrapper:
    def __init__(self, ctx: click.Context):
        self.ctx = ctx

    @cached_property
    def m(self) -> CliManager:
        from .manager import CliManager

        try:
            config = Config.read(self.ctx.params['profile']).ensure_dirs()
        except FileNotFoundError:
            config = self.ctx.invoke(configure)
        setup_logging(config, self.ctx.params['log_level'])
        _override_loop_policy()
        manager = CliManager.from_config(config)
        return manager


class _PathParam(click.Path):
    def coerce_path_result(self, value: str) -> Path:
        return Path(value)


class _EnumParam(click.Choice, Generic[_EnumT]):
    def __init__(
        self,
        choice_enum: Type[_EnumT],
        excludes: Set[_EnumT] = set(),
        case_sensitive: bool = True,
    ) -> None:
        self.choice_enum = choice_enum
        super().__init__(
            choices=[c.name for c in choice_enum if c not in excludes],
            case_sensitive=case_sensitive,
        )

    def convert(self, value: str, param: O[click.Parameter], ctx: O[click.Context]) -> _EnumT:
        parent_result = super().convert(value, param, ctx)
        return self.choice_enum[parent_result]


def _callbackify(fn: Callable[..., _R]) -> Callable[[click.Context, click.Parameter, Any], _R]:
    return lambda c, _, v: fn(c.obj.m, v)


def _show_version(ctx: click.Context, _param: click.Parameter, value: bool):
    if value:
        __version__ = get_version()
        click.echo(f'instawow, version {__version__}')
        ctx.exit()


@click.group(context_settings={'help_option_names': ('-h', '--help')})
@click.option(
    '--version',
    is_flag=True,
    default=False,
    expose_value=False,
    is_eager=True,
    callback=_show_version,
    help='Show the version and exit.',
)
@click.option(
    '--debug',
    'log_level',
    is_flag=True,
    default=False,
    callback=lambda _, __, v: 'DEBUG' if v else 'INFO',
    help='Log more things.',
)
@click.option('--profile', '-p', default='__default__', help='Activate the specified profile.')
@click.pass_context
def main(ctx: click.Context, log_level: bool, profile: str):
    "Add-on manager for World of Warcraft."
    if not ctx.obj:
        ctx.obj = ManagerWrapper(ctx)


@overload
def parse_into_defn(manager: CliManager, value: str, *, raise_invalid: bool = True) -> Defn:
    ...


@overload
def parse_into_defn(
    manager: CliManager, value: List[str], *, raise_invalid: bool = True
) -> List[Defn]:
    ...


def parse_into_defn(
    manager: CliManager, value: Union[str, List[str]], *, raise_invalid: bool = True
) -> Union[Defn, List[Defn]]:
    if not isinstance(value, str):
        defns = (parse_into_defn(manager, v, raise_invalid=raise_invalid) for v in value)
        return uniq(defns)

    pair = manager.pair_uri(value)
    if not pair:
        if raise_invalid:
            raise click.BadParameter(value)

        pair = '*', value
    source, name = pair
    return Defn(source=source, name=name)


def parse_into_defn_with_strategy(
    manager: CliManager, value: Sequence[Tuple[Strategies, str]]
) -> List[Defn]:
    defns = parse_into_defn(manager, [d for _, d in value])
    return list(map(lambda d, s: d.with_(strategy=s), defns, (s for s, _ in value)))


def parse_into_defn_with_version(
    manager: CliManager, value: Sequence[Tuple[str, str]]
) -> List[Defn]:
    defns = parse_into_defn(manager, [d for _, d in value])
    return list(map(Defn.with_version, defns, (v for v, _ in value)))


def parse_into_defn_from_json_file(manager: CliManager, path: Path) -> List[Defn]:
    faux_pkgs = models.MultiPkgModel.parse_file(path, encoding='utf-8')
    return list(map(Defn.from_pkg, cast('List[models.Pkg]', faux_pkgs.__root__)))


def _combine_into(param_name: str, fn: Callable[[CliManager, Any], List[Any]]):
    def combine(ctx: click.Context, _param: click.Parameter, value: Any):
        # Add param by ``param_name`` to param dict regardless of output
        param = ctx.params.setdefault(param_name, [])
        if value:
            param.extend(fn(ctx.obj.m, value))

    return combine


excluded_strategies = {Strategies.default, Strategies.version}


@main.command()
@click.argument(
    'addons', nargs=-1, callback=_combine_into('addons', parse_into_defn), expose_value=False
)
@click.option(
    '--import',
    '-i',
    type=_PathParam(dir_okay=False, exists=True),
    expose_value=False,
    callback=_combine_into('addons', parse_into_defn_from_json_file),
    help='Install add-ons from the output of `list -f json`.',
)
@click.option(
    '--with-strategy',
    '-s',
    multiple=True,
    type=(_EnumParam(Strategies, excluded_strategies), str),
    expose_value=False,
    callback=_combine_into('addons', parse_into_defn_with_strategy),
    metavar='<STRATEGY ADDON>...',
    help='A strategy followed by an add-on definition.  '
    'The strategies are: '
    f'{", ".join(s.name for s in Strategies if s not in excluded_strategies)}.',
)
@click.option(
    '--version',
    multiple=True,
    type=(str, str),
    expose_value=False,
    callback=_combine_into('addons', parse_into_defn_with_version),
    metavar='<VERSION ADDON>...',
    help='A version followed by an add-on definition.',
)
@click.option('--replace', '-o', is_flag=True, default=False, help='Replace existing add-ons.')
@click.pass_obj
def install(obj: ManagerWrapper, addons: Sequence[Defn], replace: bool) -> None:
    "Install add-ons."
    if not addons:
        raise click.UsageError(
            'You must provide at least one of "ADDONS", "--with-strategy", '
            '"--version" or "--import"'
        )

    results = obj.m.run(obj.m.install(addons, replace))
    Report(results.items()).generate_and_exit()


@main.command()
@click.argument('addons', nargs=-1, callback=_callbackify(parse_into_defn))
@click.pass_obj
def update(obj: ManagerWrapper, addons: Sequence[Defn]) -> None:
    "Update installed add-ons."

    def filter_results(result: E.ManagerResult):
        # Hide packages from output if they are up to date
        # and ``update`` was invoked without args
        return cast(bool, addons) or not isinstance(result, E.PkgUpToDate)

    results = obj.m.run(
        obj.m.update(addons or list(map(Defn.from_pkg, obj.m.database.query(models.Pkg).all())))
    )
    Report(results.items(), filter_results).generate_and_exit()


@main.command()
@click.argument('addons', nargs=-1, required=True, callback=_callbackify(parse_into_defn))
@click.pass_obj
def remove(obj: ManagerWrapper, addons: Sequence[Defn]) -> None:
    "Remove add-ons."
    results = obj.m.run(obj.m.remove(addons))
    Report(results.items()).generate_and_exit()


@main.command()
@click.argument('addon', callback=_callbackify(parse_into_defn))
@click.option(
    '--undo',
    is_flag=True,
    default=False,
    help='Undo rollback by reinstalling an add-on using the default strategy.',
)
@click.pass_context
def rollback(ctx: click.Context, addon: Defn, undo: bool) -> None:
    "Roll an add-on back to an older version."
    from .prompts import Choice, select

    manager: CliManager = ctx.obj.m
    limit = 10

    pkg = manager.get_pkg(addon)
    if not pkg:
        Report([(addon, E.PkgNotInstalled())]).generate_and_exit()
        return  # noop

    resolver = manager.resolvers[pkg.source]
    if not resolver.supports_rollback:
        Report(
            [(addon, E.PkgFileUnavailable('source does not support rollback'))]
        ).generate_and_exit()

    if undo:
        Report(
            chain(
                manager.run(manager.remove([addon])).items(),
                manager.run(manager.install([addon], replace=False)).items(),
            )
        ).generate_and_exit()

    versions = (
        manager.database.query(models.PkgVersionLog)
        .filter(
            models.PkgVersionLog.pkg_source == pkg.source, models.PkgVersionLog.pkg_id == pkg.id
        )
        .order_by(models.PkgVersionLog.install_time.desc())
        .limit(limit)
        .all()
    )
    if len(versions) <= 1:
        Report([(addon, E.PkgFileUnavailable('cannot find older versions'))]).generate_and_exit()

    reconstructed_defn = Defn.from_pkg(pkg)
    choices = [
        Choice(
            [('', v.version)],  # type: ignore
            disabled='installed version' if v.version == pkg.version else None,
        )
        for v in versions
    ]
    selection = select(
        f'Select version of {reconstructed_defn} for rollback', choices
    ).unsafe_ask()
    Report(
        chain(
            manager.run(manager.remove([reconstructed_defn])).items(),
            manager.run(
                manager.install([reconstructed_defn.with_version(selection)], replace=False)
            ).items(),
        )
    ).generate_and_exit()


@main.command()
@click.option(
    '--auto', '-a', is_flag=True, default=False, help='Do not ask for user confirmation.'
)
@click.option(
    '--list-unreconciled', is_flag=True, default=False, help='List unreconciled add-ons and exit.'
)
@click.pass_context
def reconcile(ctx: click.Context, auto: bool, list_unreconciled: bool) -> None:
    "Reconcile pre-installed add-ons."
    from .matchers import AddonFolder, get_folders, match_dir_names, match_toc_ids, match_toc_names
    from .models import is_pkg
    from .prompts import PkgChoice, confirm, select, skip

    preamble = dedent(
        '''\
        Use the arrow keys to navigate, <o> to open an add-on in your browser,
        enter to make a selection and <s> to skip to the next item.

        Versions that differ from the installed version or differ between
        choices are highlighted in purple.

        The reconciler will perform three passes in decreasing order of accuracy,
        looking to match source IDs and add-on names in TOC files, and folders.

        Selected add-ons will be reinstalled.

        You can also run `reconcile` in promptless mode by passing
        the `--auto` flag.  In this mode, add-ons will be reconciled
        without user input.
        '''
    )

    manager: CliManager = ctx.obj.m

    def prompt_one(addons: List[AddonFolder], pkgs: List[models.Pkg]) -> Union[Defn, Tuple[()]]:
        def construct_choice(pkg: models.Pkg):
            defn = Defn.from_pkg(pkg)
            title = [
                ('', str(defn)),
                ('', '=='),
                ('class:highlight-sub' if highlight_version else '', pkg.version),
            ]
            return PkgChoice(title, defn, pkg=pkg)

        # Highlight version if there's multiple of them
        highlight_version = len({i.version for i in chain(addons, pkgs)}) > 1
        choices = list(chain(map(construct_choice, pkgs), (skip,)))
        addon = addons[0]
        # Using 'unsafe_ask' to let ^C bubble up
        selection = select(f'{addon.name} [{addon.version or "?"}]', choices).unsafe_ask()
        return selection

    def prompt(groups: Sequence[Tuple[List[AddonFolder], List[Defn]]]) -> Iterable[Defn]:
        uniq_defns = uniq(d for _, b in groups for d in b)
        results = manager.run(manager.resolve(uniq_defns))
        for addons, defns in groups:
            shortlist = list(filter(is_pkg, (results[d] for d in defns)))
            if shortlist:
                if auto:
                    selection = Defn.from_pkg(shortlist[0])
                else:
                    selection = prompt_one(addons, shortlist)
                selection and (yield selection)

    def match_all() -> Generator[List[Defn], FrozenSet[AddonFolder], None]:
        # Match in order of increasing heuristicitivenessitude
        for fn in (
            match_toc_ids,
            match_dir_names,
            match_toc_names,
        ):
            groups = manager.run(fn(manager, (yield [])))
            yield list(prompt(groups))

    leftovers = get_folders(manager)
    if list_unreconciled:
        table_rows = [('unreconciled',), *((f.name,) for f in sorted(leftovers))]
        click.echo(tabulate(table_rows))
        return
    elif not leftovers:
        click.echo('No add-ons left to reconcile.')
        return

    if not auto:
        click.echo(preamble)

    matcher = match_all()
    for _ in matcher:  # Skip over consumer yields
        selections = matcher.send(leftovers)
        if selections and (auto or confirm('Install selected add-ons?').unsafe_ask()):
            results = manager.run(manager.install(selections, replace=True))
            Report(results.items()).generate()

        leftovers = get_folders(manager)

    if leftovers:
        click.echo()
        table_rows = [('unreconciled',), *((f.name,) for f in sorted(leftovers))]
        click.echo(tabulate(table_rows))


@main.command()
@click.argument('search-terms', nargs=-1, required=True, callback=lambda _, __, v: ' '.join(v))
@click.option(
    '--limit',
    '-l',
    default=10,
    type=click.IntRange(1, 20, clamp=True),
    help='A number to limit results to.',
)
@click.pass_context
def search(ctx: click.Context, search_terms: str, limit: int) -> None:
    "Search for add-ons to install."
    from .prompts import PkgChoice, checkbox, confirm

    manager: CliManager = ctx.obj.m

    pkgs = manager.run(manager.search(search_terms, limit))
    if pkgs:
        choices = [PkgChoice(f'{p.name}  ({d}=={p.version})', d, pkg=p) for d, p in pkgs.items()]
        selections = checkbox('Select add-ons to install', choices=choices).unsafe_ask()
        if selections and confirm('Install selected add-ons?').unsafe_ask():
            ctx.invoke(install, addons=selections)


class ListFormats(enum.Enum):
    simple = 'simple'
    detailed = 'detailed'
    json = 'json'


@main.command('list')
@click.argument(
    'addons', nargs=-1, callback=_callbackify(partial(parse_into_defn, raise_invalid=False))
)
@click.option(
    '--format',
    '-f',
    'output_format',
    type=_EnumParam(ListFormats),
    default=ListFormats.simple.name,
    show_default=True,
    help='Change the output format.',
)
@click.pass_obj
def list_installed(
    obj: ManagerWrapper, addons: Sequence[Defn], output_format: ListFormats
) -> None:
    "List installed add-ons."
    from sqlalchemy import and_, or_

    def format_deps(pkg: models.Pkg):
        deps = (Defn(source=pkg.source, name=d.id) for d in pkg.deps)
        deps = (d.with_(name=getattr(obj.m.get_pkg(d), 'slug', d.name)) for d in deps)
        return map(str, deps)

    def get_wowi_desc_from_toc(pkg: models.Pkg):
        if pkg.source == 'wowi':
            toc_reader = TocReader.from_parent_folder(obj.m.config.addon_dir / pkg.folders[0].name)
            return toc_reader['Notes'].value
        else:
            return pkg.description

    pkgs = (
        obj.m.database.query(models.Pkg)
        .filter(
            or_(
                *(
                    models.Pkg.slug.contains(d.name)
                    if d.source == '*'
                    else and_(
                        models.Pkg.source == d.source,
                        or_(models.Pkg.id == d.name, models.Pkg.slug == d.name),
                    )
                    for d in addons
                )
            )
        )
        .order_by(models.Pkg.source, models.Pkg.name)
        .all()
    )
    if output_format is ListFormats.json:
        click.echo(models.MultiPkgModel.parse_obj(pkgs).json(indent=2))
    elif output_format is ListFormats.detailed:
        formatter = click.HelpFormatter(max_width=99)
        for pkg in pkgs:
            with formatter.section(Defn.from_pkg(pkg)):
                formatter.write_dl(
                    (
                        ('Name', pkg.name),
                        ('Description', get_wowi_desc_from_toc(pkg)),
                        ('URL', pkg.url),
                        ('Version', pkg.version),
                        ('Date published', pkg.date_published.isoformat(' ', 'minutes')),
                        ('Folders', ', '.join(f.name for f in pkg.folders)),
                        ('Dependencies', ', '.join(format_deps(pkg))),
                        ('Options', f'{{"strategy": "{pkg.options.strategy}"}}'),
                    )
                )
        click.echo(formatter.getvalue(), nl=False)
    else:
        click.echo(''.join(f'{Defn.from_pkg(p)}\n' for p in pkgs), nl=False)


@main.command(hidden=True)
@click.argument('addon', callback=_callbackify(partial(parse_into_defn, raise_invalid=False)))
@click.pass_context
def info(ctx: click.Context, addon: Defn) -> None:
    "Alias of `list -f detailed`."
    ctx.invoke(list_installed, addons=(addon,), output_format=ListFormats.detailed)


@main.command()
@click.argument('addon', callback=_callbackify(partial(parse_into_defn, raise_invalid=False)))
@click.pass_obj
def reveal(obj: ManagerWrapper, addon: Defn) -> None:
    "Bring an add-on up in your file manager."
    pkg = obj.m.get_pkg(addon, partial_match=True)
    if pkg:
        import webbrowser

        webbrowser.open((obj.m.config.addon_dir / pkg.folders[0].name).as_uri())
    else:
        Report([(addon, E.PkgNotInstalled())]).generate_and_exit()


def _show_active_config(ctx: click.Context, _param: click.Parameter, value: bool):
    if value:
        click.echo(ctx.obj.m.config.json(indent=2))
        ctx.exit()


@main.command()
@click.option(
    '--active',
    'show_active',
    is_flag=True,
    default=False,
    expose_value=False,
    is_eager=True,
    callback=_show_active_config,
    help='Show the active configuration and exit.',
)
@click.pass_context
def configure(ctx: click.Context) -> Config:
    "Configure instawow."
    from prompt_toolkit.completion import PathCompleter, WordCompleter
    from questionary import text

    from .prompts import PydanticValidator

    profile = ctx.find_root().params['profile']
    game_flavours = Config.__fields__['game_flavour'].type_.__args__

    addon_dir = text(
        'Add-on directory: ',
        completer=PathCompleter(only_directories=True, expanduser=True),
        validate=PydanticValidator(Config, 'addon_dir'),
    ).unsafe_ask()
    game_flavour = text(
        'Game flavour: ',
        default='classic' if '_classic_' in addon_dir else 'retail',
        completer=WordCompleter(game_flavours),
        validate=PydanticValidator(Config, 'game_flavour'),
    ).unsafe_ask()
    config = Config(profile=profile, addon_dir=addon_dir, game_flavour=game_flavour).write()
    click.echo(f'Configuration written to: {config.config_file}')
    return config


@main.group('weakauras-companion')
def _weakauras_group():
    "Manage your WeakAuras."


@_weakauras_group.command('build')
@click.option(
    '--account',
    '-a',
    required=True,
    help='Your account name.  This is used to locate the WeakAuras data file.',
)
@click.pass_obj
def build_weakauras_companion(obj: ManagerWrapper, account: str) -> None:
    "Build the WeakAuras Companion add-on."
    from .wa_updater import BuilderConfig, WaCompanionBuilder

    config = BuilderConfig(account=account)
    obj.m.run(WaCompanionBuilder(obj.m, config).build())


@_weakauras_group.command('list')
@click.option(
    '--account',
    '-a',
    required=True,
    help='Your account name.  This is used to locate the WeakAuras data file.',
)
@click.pass_obj
def list_installed_wago_auras(obj: ManagerWrapper, account: str) -> None:
    "List WeakAuras installed from Wago."
    from textwrap import fill

    from .wa_updater import BuilderConfig, WaCompanionBuilder

    config = BuilderConfig(account=account)
    aura_groups = WaCompanionBuilder(obj.m, config).extract_installed_auras()
    installed_auras = sorted(
        (g.Meta.model.__name__, fill(a.id, width=30, max_lines=1), a.url)
        for g in aura_groups
        for v in g.entries.values()
        for a in v
        if not a.parent
    )
    click.echo(tabulate([('type', 'name', 'URL'), *installed_auras]))


@main.command(hidden=True)
@click.argument('filename', type=_PathParam(dir_okay=False))
def generate_catalogue(filename: Path) -> None:
    "Generate the master catalogue."
    import asyncio

    from .resolvers import MasterCatalogue

    catalogue = asyncio.run(MasterCatalogue.collate())
    filename.write_text(catalogue.json(indent=2), encoding='utf-8')
    filename.with_suffix(f'.compact{filename.suffix}').write_text(
        catalogue.json(separators=(',', ':')), encoding='utf-8'
    )


@main.command(hidden=True)
@click.pass_context
def listen(ctx: click.Context) -> None:
    "Fire up the WebSocket server."
    import asyncio

    from .json_rpc_server import listen

    dummy_config = Config.get_dummy_config(profile='__jsonrpc__').ensure_dirs()
    log_level = ctx.find_root().params['log_level']
    setup_logging(dummy_config, log_level)
    _override_loop_policy()
    asyncio.run(listen())



import pytest

from instawow import exceptions as E
from instawow.models import Pkg
from instawow.resolvers import Defn, Strategies


@pytest.fixture(autouse=True)
def mock(mock_all):
    pass


@pytest.mark.asyncio
@pytest.mark.parametrize(
    'strategy', [Strategies.default, Strategies.latest, Strategies.any_flavour]
)
async def test_resolve_curse_simple_pkgs(manager, request, strategy):
    results = await manager.resolve(
        [
            Defn(source='curse', name='tomcats', strategy=strategy),
            Defn(source='curse', name='mythic-dungeon-tools', strategy=strategy),
            Defn(source='curse', name='classiccastbars', strategy=strategy),
            Defn(source='curse', name='elkbuffbars', strategy=strategy),
        ]
    )
    separate, retail_only, classic_only, flavour_explosion = results.values()

    assert isinstance(separate, Pkg)
    if manager.config.is_classic:
        if strategy is Strategies.any_flavour:
            assert 'classic' not in separate.version
            assert isinstance(retail_only, Pkg)
        else:
            assert 'classic' in separate.version
            assert (
                isinstance(retail_only, E.PkgFileUnavailable)
                and retail_only.message
                == f"no files compatible with classic using '{strategy.name}' strategy"
            )
        assert isinstance(classic_only, Pkg)
    else:
        assert 'classic' not in separate.version
        assert isinstance(retail_only, Pkg)
        if strategy is Strategies.any_flavour:
            assert isinstance(classic_only, Pkg)
        else:
            assert (
                isinstance(classic_only, E.PkgFileUnavailable)
                and classic_only.message
                == f"no files compatible with retail using '{strategy.name}' strategy"
            )

    versions = {*request.config.cache.get('flavour_explosion', ()), flavour_explosion.version}
    assert len(versions) == 1
    request.config.cache.set('flavour_explosion', tuple(versions))


@pytest.mark.asyncio
async def test_resolve_curse_latest_pkg(manager):
    (latest_pkg,) = (
        await manager.resolve([Defn(source='curse', name='tomcats', strategy=Strategies.latest)])
    ).values()
    assert latest_pkg.options.strategy == 'latest'


@pytest.mark.asyncio
async def test_resolve_curse_versioned_pkg(manager):
    (versioned_pkg,) = (
        await manager.resolve([Defn.get('curse', 'molinari').with_version('70300.51-Release')])
    ).values()
    assert (
        versioned_pkg.options.strategy == 'version' and versioned_pkg.version == '70300.51-Release'
    )


@pytest.mark.asyncio
async def test_resolve_curse_deps(manager):
    if manager.config.is_classic:
        pytest.skip('no classic equivalent')

    defns = [Defn(source='curse', name='mechagon-rare-share', strategy=Strategies.default)]
    with_deps = await manager.resolve(defns, with_deps=True)
    assert ['mechagon-rare-share', 'rare-share'] == [d.slug for d in with_deps.values()]


@pytest.mark.asyncio
async def test_resolve_wowi_pkgs(manager):
    results = await manager.resolve(
        [
            Defn(source='wowi', name='13188-molinari'),
            Defn(source='wowi', name='13188', strategy=Strategies.latest),
        ]
    )
    either, invalid = results.values()

    assert isinstance(either, Pkg)
    assert (
        isinstance(invalid, E.PkgStrategyUnsupported)
        and invalid.message == "'latest' strategy is not valid for source"
    )


@pytest.mark.asyncio
async def test_resolve_tukui_pkgs(manager):
    results = await manager.resolve(
        [
            Defn(source='tukui', name='1'),
            Defn(source='tukui', name='-1'),
            Defn(source='tukui', name='tukui'),
            Defn(source='tukui', name='1', strategy=Strategies.latest),
        ]
    )
    either, retail_id, retail_slug, invalid = results.values()

    assert isinstance(either, Pkg)
    if manager.config.is_classic:
        assert either.name == 'Tukui'
        assert isinstance(retail_id, E.PkgNonexistent)
        assert isinstance(retail_slug, E.PkgNonexistent)
    else:
        assert either.name == 'MerathilisUI'
        assert isinstance(retail_id, Pkg) and retail_id.name == 'Tukui'
        assert isinstance(retail_slug, Pkg) and retail_slug.name == 'Tukui'
    assert (
        isinstance(invalid, E.PkgStrategyUnsupported)
        and invalid.message == "'latest' strategy is not valid for source"
    )


@pytest.mark.asyncio
async def test_resolve_github_pkgs(manager):
    results = await manager.resolve(
        [
            Defn(source='github', name='AdiAddons/AdiButtonAuras'),
            Defn(
                source='github',
                name='AdiAddons/AdiButtonAuras',
                strategy=Strategies.version,
                strategy_vals=('2.1.0',),
            ),
            Defn(
                source='github',
                name='AdiAddons/AdiButtonAuras',
                strategy=Strategies.version,
                strategy_vals=('2.0.19',),
            ),
            Defn(source='github', name='WeakAuras/WeakAuras2'),
            Defn(source='github', name='p3lim-wow/Molinari'),
            Defn(source='github', name='layday/foo-bar'),
        ]
    )
    (
        lib_and_nolib,
        older_version,
        assetless,
        retail_and_classic,
        releaseless,
        missing,
    ) = results.values()
    assert 'nolib' not in lib_and_nolib.download_url
    assert older_version.options.strategy == 'version' and older_version.version == '2.1.0'
    assert isinstance(assetless, E.PkgFileUnavailable)
    assert (
        isinstance(releaseless, E.PkgFileUnavailable)
        and releaseless.message == 'release not found'
    )
    assert ('classic' in retail_and_classic.download_url) is manager.config.is_classic
    assert isinstance(missing, E.PkgNonexistent)



from pathlib import Path

import pytest

from instawow.manager import find_zip_base_dirs, make_zip_member_filter
from instawow.utils import TocReader, bucketise, merge_intersecting_sets, tabulate


@pytest.fixture
def fake_addon():
    yield Path(__file__).parent / 'fixtures' / 'FakeAddon'


def test_find_zip_base_dirs_can_find_explicit_dirs():
    assert find_zip_base_dirs(['b/', 'b/b.toc']) == {'b'}


def test_find_zip_base_dirs_can_find_implicit_dirs():
    assert find_zip_base_dirs(['b/b.toc']) == {'b'}


def test_find_zip_base_dirs_discards_files_in_root():
    assert find_zip_base_dirs(['a', 'b/b.toc', 'c/']) == {'b', 'c'}


def test_make_zip_member_filter_discards_names_with_prefix_not_in_dirs():
    is_member = make_zip_member_filter({'b'})
    assert list(map(is_member, ['a/', 'b/', 'aa/', 'bb/'])) == [False, True, False, False]


def test_loading_toc_from_path(fake_addon):
    TocReader.from_path(fake_addon / 'FakeAddon.toc')
    with pytest.raises(FileNotFoundError):
        TocReader.from_path(fake_addon / 'MissingToc.toc')

    TocReader.from_parent_folder(fake_addon)
    with pytest.raises(FileNotFoundError):
        TocReader.from_parent_folder(fake_addon.parent / 'MissingAddon')


def test_parsing_toc_entries(fake_addon):
    toc_reader = TocReader.from_parent_folder(fake_addon)
    assert toc_reader.entries == {
        'Normal': 'Normal entry',
        'Compact': 'Compact entry',
    }


def test_indexing_toc_entries(fake_addon):
    toc_reader = TocReader.from_parent_folder(fake_addon)
    assert toc_reader['Normal'] == ('Normal', 'Normal entry')
    assert toc_reader['Compact'] == ('Compact', 'Compact entry')
    assert toc_reader['Indented'] == ('Indented', '')
    assert toc_reader['Comment'] == ('Comment', '')
    assert toc_reader['Nonexistent'] == ('Nonexistent', '')


def test_multiindexing_toc_entries(fake_addon):
    toc_reader = TocReader.from_parent_folder(fake_addon)
    assert toc_reader['Normal', 'Compact'] == ('Normal', 'Normal entry')
    assert toc_reader['Compact', 'Normal'] == ('Compact', 'Compact entry')
    assert toc_reader['Indented', 'Normal'] == ('Normal', 'Normal entry')
    assert toc_reader['Nonexistent', 'Indented', 'Normal'] == ('Normal', 'Normal entry')


def test_bucketise_bucketises_by_putting_things_in_a_bucketing_bucket():
    assert bucketise(iter([1, 1, 0, 1]), bool) == {True: [1, 1, 1], False: [0]}


def test_tabulate(fake_addon):
    toc_reader = TocReader.from_parent_folder(fake_addon)
    data = [('key', 'value'), *toc_reader.entries.items()]
    assert (
        tabulate(data)
        == '''\
  key        value    \n\
-------  -------------
Normal   Normal entry \n\
Compact  Compact entry\
'''
    )


def test_merge_intersecting_sets_in_noncontiguous_collection():
    collection = [
        {'a'},
        {'b', 'c'},
        {'a', 'd'},
        {'e'},
        {'c', 'f'},
        {'g', 'a'},
    ]
    output = [
        {'a', 'd', 'g'},
        {'b', 'c', 'f'},
        {'e'},
    ]
    assert sorted(merge_intersecting_sets(collection)) == output



from __future__ import annotations

from contextlib import suppress
from functools import total_ordering
from pathlib import Path
import re
from typing import TYPE_CHECKING, FrozenSet, List, Tuple, cast

from .models import PkgFolder
from .resolvers import CurseResolver, Defn, TukuiResolver, WowiResolver
from .utils import TocReader, bucketise, cached_property, merge_intersecting_sets, uniq

if TYPE_CHECKING:
    from .manager import Manager

    MatchGroups = List[Tuple[List['AddonFolder'], List[Defn]]]


_ids_to_sources = {
    'X-Curse-Project-ID': CurseResolver.source,
    'X-Tukui-ProjectID': TukuiResolver.source,
    'X-WoWI-ID': WowiResolver.source,
}
_sources_to_sort_weights = {
    CurseResolver.source: 0,
    TukuiResolver.source: 2,
    WowiResolver.source: 1,
}


@total_ordering
class AddonFolder:
    def __init__(self, name: str, toc_reader: TocReader) -> None:
        self.name = name
        self.toc_reader = toc_reader

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self.name}>'

    @cached_property
    def defns_from_toc(self) -> FrozenSet[Defn]:
        return frozenset(
            Defn(source=s, name=i)
            for s, i in ((s, self.toc_reader[k].value) for k, s in _ids_to_sources.items())
            if i
        )

    @cached_property
    def version(self) -> str:
        return self.toc_reader['Version', 'X-Packaged-Version', 'X-Curse-Packaged-Version'].value

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (self.__class__, str)):
            return NotImplemented
        return self.name == other

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, (self.__class__, str)):
            return NotImplemented
        return self.name < other


def get_folders(manager: Manager, exclude_own: bool = True) -> FrozenSet[AddonFolder]:
    def make_addon_folder(path: Path):
        if path.name not in own_folders and path.is_dir() and not path.is_symlink():
            with suppress(FileNotFoundError):
                return AddonFolder(path.name, TocReader.from_parent_folder(path))

    if exclude_own:
        own_folders = {f.name for f in manager.database.query(PkgFolder).all()}
    else:
        own_folders = set()
    return frozenset(f for f in map(make_addon_folder, manager.config.addon_dir.iterdir()) if f)


async def match_toc_ids(manager: Manager, leftovers: FrozenSet[AddonFolder]) -> MatchGroups:
    "Attempt to match add-ons from TOC-file source ID entries."

    def bucket_keyer(value: AddonFolder):
        return next(d for d in defns if value.defns_from_toc & d)

    def sort_keyer(value: Defn):
        return _sources_to_sort_weights[value.source]

    matches = [a for a in sorted(leftovers) if a.defns_from_toc]
    defns = list(merge_intersecting_sets(a.defns_from_toc for a in matches))
    return [(f, sorted(b, key=sort_keyer)) for b, f in bucketise(matches, bucket_keyer).items()]


async def match_dir_names(manager: Manager, leftovers: FrozenSet[AddonFolder]) -> MatchGroups:
    "Attempt to match folders against the master catalogue."

    def bucket_keyer(value: Tuple[FrozenSet[AddonFolder], Defn]):
        return next(f for f in folders if value[0] & f)

    def sort_keyer(value: Tuple[FrozenSet[AddonFolder], Defn]):
        folders, defn = value
        return (-len(folders), _sources_to_sort_weights[defn.source])

    await manager.synchronise()

    # We can't use an intersection here because it's not guaranteed to
    # return ``AddonFolder``s - the duck typing semantics of '&' appear
    # to be undefined
    matches = [
        (
            frozenset(e for e in leftovers if e in cast('List[AddonFolder]', f)),
            Defn(source=i.source, name=i.id),
        )
        for i in manager.catalogue.__root__
        for f in i.folders
        if manager.config.game_flavour in i.compatibility and frozenset(f) <= leftovers
    ]
    folders = list(merge_intersecting_sets(f for f, _ in matches))
    return [
        (sorted(f), uniq(d for _, d in sorted(b, key=sort_keyer)))
        for f, b in bucketise(matches, bucket_keyer).items()
    ]


async def match_toc_names(manager: Manager, leftovers: FrozenSet[AddonFolder]) -> MatchGroups:
    "Attempt to match add-ons from TOC-file name entries."

    def normalise(value: str):
        return re.sub(r'[^0-9A-Za-z]', '', value.casefold())

    await manager.synchronise()

    norm_to_items = bucketise(manager.catalogue.__root__, key=lambda i: normalise(i.name))
    matches = ((e, norm_to_items.get(normalise(e.name))) for e in sorted(leftovers))
    return [([e], uniq(Defn(source=i.source, name=i.id) for i in m)) for e, m in matches if m]



import pytest

from instawow import exceptions as E
from instawow.resolvers import Defn, Strategies


@pytest.mark.asyncio
async def test_finding_damaged_pkgs(mock_all, manager):
    molinari_defn = Defn.get('curse', 'molinari')
    install_result = await manager.install([molinari_defn], False)
    installed_pkg = install_result[molinari_defn].pkg
    assert not manager.find_damaged_pkgs()
    (manager.config.addon_dir / installed_pkg.folders[0].name).rename(
        manager.config.addon_dir / 'Foo'
    )
    damaged_pkgs = manager.find_damaged_pkgs()
    assert installed_pkg in damaged_pkgs


@pytest.mark.asyncio
async def test_pinning_supported(mock_all, manager):
    defn = Defn.get('curse', 'molinari')
    install_result = await manager.install([defn], False)
    pkg = install_result[defn].pkg
    version = pkg.version

    for new_defn in (defn.with_version(pkg.version), defn):
        pin_result = await manager.pin([new_defn])
        pinned_pkg = pin_result[new_defn].pkg
        assert pkg.options.strategy == pinned_pkg.options.strategy == new_defn.strategy.name
        assert version == pinned_pkg.version


@pytest.mark.asyncio
async def test_pinning_unsupported(mock_all, manager):
    molinari_defn = Defn.get('wowi', '13188')
    await manager.install([molinari_defn], False)
    installed_pkg = manager.get_pkg(molinari_defn)
    assert installed_pkg.options.strategy == 'default'
    result = await manager.pin([molinari_defn])
    assert (
        isinstance(result[molinari_defn], E.PkgStrategyUnsupported)
        and result[molinari_defn].strategy == Strategies.version
    )
    assert installed_pkg.options.strategy == 'default'


@pytest.mark.asyncio
async def test_pinning_nonexistent(mock_all, manager):
    molinari_defn = Defn.get('wowi', '13188')
    result = await manager.pin([molinari_defn])
    assert isinstance(result[molinari_defn], E.PkgNotInstalled)



from aiohttp.test_utils import TestClient, TestServer
from aiohttp_rpc import JsonRpcRequest as Request, JsonRpcResponse as Response
import pytest

from instawow.config import Config
from instawow.json_rpc_server import create_app, serialise_response


@pytest.fixture
async def ws(full_config, monkeypatch):
    monkeypatch.setenv('INSTAWOW_CONFIG_DIR', str(full_config['config_dir']))
    app = await create_app()
    server = TestServer(app)
    async with TestClient(server) as client, client.ws_connect('/v0') as ws:
        yield ws


@pytest.mark.asyncio
async def test_write_config(request, full_config, ws):
    config_values = {**full_config, 'profile': request.node.name}
    rpc_request = Request(
        method='config.write',
        params={'values': config_values},
        msg_id=request.node.name,
    )
    await ws.send_str(serialise_response(rpc_request.to_dict()))
    rpc_response = Response.from_dict(await ws.receive_json())
    assert rpc_response.msg_id == request.node.name
    assert Config.parse_obj(rpc_response.result) == Config.parse_obj(config_values)


@pytest.mark.asyncio
async def test_write_config_with_invalid_params(request, full_config, ws):
    rpc_request = Request(
        method='config.write',
        params={'values': {**full_config, 'game_flavour': 'strawberry'}},
        msg_id=request.node.name,
    )
    await ws.send_str(serialise_response(rpc_request.to_dict()))
    rpc_response = Response.from_dict(await ws.receive_json())
    assert rpc_response.msg_id == request.node.name
    assert rpc_response.error
    error = rpc_response.error
    assert error.code == -32001
    assert rpc_response.error.message == 'invalid configuration parameters'
    assert rpc_response.error.data == [
        {
            'loc': ['game_flavour'],
            'msg': "unexpected value; permitted: 'retail', 'classic'",
            'type': 'value_error.const',
            'ctx': {'given': 'strawberry', 'permitted': ['retail', 'classic']},
        }
    ]


@pytest.mark.asyncio
async def test_install_with_invalid_params(request, ws):
    rpc_request = Request(
        method='install',
        params={},
        msg_id=request.node.name,
    )
    await ws.send_str(serialise_response(rpc_request.to_dict()))
    rpc_response = Response.from_dict(await ws.receive_json())
    assert rpc_response.error and rpc_response.error.code == -32602


@pytest.mark.xfail
@pytest.mark.asyncio
async def test_install_with_uninitialised_profile(request, ws):
    rpc_request = Request(
        method='install',
        params={
            'profile': request.node.name,
            'defns': [{'source': 'curse', 'name': 'molinari'}],
            'replace': False,
        },
        msg_id=request.node.name,
    )
    await ws.send_str(serialise_response(rpc_request.to_dict()))
    rpc_response = Response.from_dict(await ws.receive_json())
    assert rpc_response.error and rpc_response.error.code == -32001



from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Any, List, Optional, Sequence, Tuple, Type

from prompt_toolkit.application import Application
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys
from prompt_toolkit.styles import Style
from prompt_toolkit.validation import ValidationError, Validator
import pydantic
from questionary import Choice, confirm as _confirm
from questionary.prompts.common import InquirerControl, Separator, create_inquirer_layout
from questionary.question import Question

if TYPE_CHECKING:
    from prompt_toolkit.document import Document

    from .models import Pkg


class PydanticValidator(Validator):
    "One-off validators for Pydantic model fields."

    def __init__(self, model: Type[pydantic.BaseModel], field: str) -> None:
        self.model = model
        self.field = field

    def validate(self, document: Document) -> None:
        try:
            self.model.parse_obj({self.field: document.text})
        except pydantic.ValidationError as error:
            error_at_loc = next((e for e in error.errors() if e['loc'] == (self.field,)), None)
            if error_at_loc:
                raise ValidationError(0, error_at_loc['msg'])


class PkgChoice(Choice):
    def __init__(self, *args: Any, pkg: Optional[Pkg] = None, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.pkg = pkg


qstyle = Style(
    [
        ('qmark', 'fg:ansicyan'),
        ('answer', 'fg: nobold'),
        ('highlight-sub', 'fg:ansimagenta'),
        ('skipped', 'fg:ansiyellow'),
        ('question', 'nobold'),
        ('x-question', 'bold'),
    ]
)

skip = Choice([('', 'skip')], ())  # type: ignore

confirm = partial(_confirm, style=qstyle)


def checkbox(message: str, choices: Sequence[Choice], **prompt_kwargs: Any) -> Question:
    # This is a cut-down version of ``questionary.checkbox`` with the addition
    # of an <o> key binding for opening package URLs

    def get_prompt_tokens():
        tokens: List[Tuple[str, str]] = [('class:x-question', message)]
        if ic.is_answered:
            tokens = [*tokens, (('class:answer', '  done'))]
        else:
            tokens = [
                *tokens,
                (
                    (
                        'class:instruction',
                        '  (use arrow keys to move, <space> to select, <o> to view in your browser)',
                    )
                ),
            ]
        return tokens

    ic = InquirerControl(
        choices, None, use_indicator=False, use_shortcuts=False, use_pointer=True  # type: ignore
    )
    bindings = KeyBindings()

    @bindings.add(Keys.ControlQ, eager=True)
    @bindings.add(Keys.ControlC, eager=True)
    def _(event: Any):
        event.app.exit(exception=KeyboardInterrupt, style='class:aborting')

    @bindings.add(' ', eager=True)
    def toggle(event: Any):
        pointed_choice = ic.get_pointed_at().value
        if pointed_choice in ic.selected_options:
            ic.selected_options.remove(pointed_choice)
        else:
            ic.selected_options.append(pointed_choice)

    @bindings.add('i', eager=True)
    def invert(event: Any):
        inverted_selection = [
            c.value
            for c in ic.choices
            if not isinstance(c, Separator)
            and c.value not in ic.selected_options
            and not c.disabled
        ]
        ic.selected_options = inverted_selection

    @bindings.add(Keys.Down, eager=True)
    @bindings.add('j', eager=True)
    def move_cursor_down(event: Any):
        ic.select_next()
        while not ic.is_selection_valid():
            ic.select_next()

    @bindings.add(Keys.Up, eager=True)
    @bindings.add('k', eager=True)
    def move_cursor_up(event: Any):
        ic.select_previous()
        while not ic.is_selection_valid():
            ic.select_previous()

    @bindings.add(Keys.ControlM, eager=True)
    def set_answer(event: Any):
        ic.is_answered = True
        event.app.exit(result=[c.value for c in ic.get_selected_values()])

    @bindings.add('o', eager=True)
    def open_url(event: Any):
        pkg = ic.get_pointed_at().pkg
        if pkg:
            import webbrowser

            webbrowser.open(pkg.url)

    @bindings.add(Keys.Any)
    def other(event: Any):
        # Disallow inserting other text
        pass

    layout = create_inquirer_layout(ic, get_prompt_tokens, **prompt_kwargs)
    app = Application(layout=layout, key_bindings=bindings, style=qstyle, **prompt_kwargs)
    return Question(app)


def select(message: str, choices: Sequence[Choice], **prompt_kwargs: Any) -> Question:
    def get_prompt_tokens():
        tokens: List[Tuple[str, str]] = [('', '- '), ('class:x-question', message)]
        if ic.is_answered:
            answer = ''.join(t for _, t in ic.get_pointed_at().title)
            tokens = [*tokens, ('', '  '), ('class:skipped' if answer == 'skip' else '', answer)]
        return tokens

    ic = InquirerControl(
        choices, None, use_indicator=False, use_shortcuts=False, use_pointer=True  # type: ignore
    )
    bindings = KeyBindings()

    @bindings.add(Keys.ControlQ, eager=True)
    @bindings.add(Keys.ControlC, eager=True)
    def _(event: Any):
        event.app.exit(exception=KeyboardInterrupt, style='class:aborting')

    @bindings.add(Keys.Down, eager=True)
    @bindings.add('j', eager=True)
    def move_cursor_down(event: Any):
        ic.select_next()
        while not ic.is_selection_valid():
            ic.select_next()

    @bindings.add(Keys.Up, eager=True)
    @bindings.add('k', eager=True)
    def move_cursor_up(event: Any):
        ic.select_previous()
        while not ic.is_selection_valid():
            ic.select_previous()

    @bindings.add(Keys.ControlM, eager=True)
    def set_answer(event: Any):
        ic.is_answered = True
        event.app.exit(result=ic.get_pointed_at().value)

    @bindings.add('o', eager=True)
    def open_url(event: Any):
        pkg = getattr(ic.get_pointed_at(), 'pkg', None)
        if pkg:
            import webbrowser

            webbrowser.open(pkg.url)

    @bindings.add('s', eager=True)
    def skip(event: Any):
        ic.pointed_at = -1
        ic.is_answered = True
        event.app.exit(result=ic.get_pointed_at().value)

    @bindings.add(Keys.Any)
    def other(event: Any):
        # Disallow inserting other text
        pass

    layout = create_inquirer_layout(ic, get_prompt_tokens, **prompt_kwargs)
    app = Application(layout=layout, key_bindings=bindings, style=qstyle, **prompt_kwargs)
    return Question(app)



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Sequence, Set

if TYPE_CHECKING:
    from .models import Pkg
    from .resolvers import Strategies


class ManagerResult:
    kind: ClassVar[str]
    message_template: ClassVar[str]

    @property
    def message(self) -> str:
        return self.message_template.format(self=self)


class PkgInstalled(ManagerResult):
    kind = 'success'
    message_template = 'installed {self.pkg.version}'

    def __init__(self, pkg: Pkg) -> None:
        super().__init__()
        self.pkg = pkg


class PkgUpdated(ManagerResult):
    kind = 'success'
    message_template = 'updated {self.old_pkg.version} to {self.new_pkg.version}'

    def __init__(self, old_pkg: Pkg, new_pkg: Pkg) -> None:
        super().__init__()
        self.old_pkg = old_pkg
        self.new_pkg = new_pkg


class PkgRemoved(ManagerResult):
    kind = 'success'
    message_template = 'removed'

    def __init__(self, old_pkg: Pkg) -> None:
        super().__init__()
        self.old_pkg = old_pkg


class ManagerError(ManagerResult, Exception):
    kind = 'failure'


class PkgAlreadyInstalled(ManagerError):
    message_template = 'package already installed'


class PkgConflictsWithInstalled(ManagerError):
    message_template = 'package folders conflict with installed package {self.conflicts[0]}'

    def __init__(self, conflicts: Sequence[Pkg]) -> None:
        from .resolvers import Defn

        super().__init__()
        self.conflicts = [Defn.from_pkg(c) for c in conflicts]


class PkgConflictsWithForeign(ManagerError):
    message_template = 'package folders conflict with {self.folders}'

    def __init__(self, folders: Set[str]) -> None:
        super().__init__()
        self.folders = ', '.join(f"'{f}'" for f in folders)


class PkgNonexistent(ManagerError):
    message_template = 'package does not exist'


class PkgFileUnavailable(ManagerError):
    message_template = 'package file is not available for download'

    def __init__(self, specialised_message: Optional[str] = None) -> None:
        super().__init__()
        self.specialised_message = specialised_message

    @property
    def message(self) -> str:
        return self.specialised_message or super().message


class PkgNotInstalled(ManagerError):
    message_template = 'package is not installed'


class PkgSourceInvalid(ManagerError):
    message_template = 'package source is invalid'


class PkgUpToDate(ManagerError):
    message_template = 'package is up to date'


class PkgStrategyUnsupported(ManagerError):
    message_template = '{self.strategy.name!r} strategy is not valid for source'

    def __init__(self, strategy: Strategies) -> None:
        super().__init__()
        self.strategy = strategy


class InternalError(ManagerResult, Exception):
    kind = 'error'
    message_template = 'internal error'

    def __init__(self, error: BaseException, stringify_error: bool = False) -> None:
        super().__init__()
        self.error = error
        self.stringify_error = stringify_error

    @property
    def message(self) -> str:
        return (
            f'{self.message_template}: "{self.error}"' if self.stringify_error else super().message
        )
